# Copyright 2012 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
from io import BytesIO
import json
from muranopkgcheck import manager as check_manager
from muranopkgcheck import pkg_loader as check_pkg_loader
from muranopkgcheck import validators as check_validators
import os
import re
import shutil
import sys
import tempfile
import textwrap
import uuid
import warnings
import zipfile

from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import encodeutils
from oslo_utils import uuidutils

import prettytable
import requests
import urllib
import yaml
import yaql

from muranoclient.common import exceptions
from muranoclient.i18n import _

try:
    import yaql.language  # noqa

    from muranoclient.common.yaqlexpression import YaqlExpression
except ImportError:
    # no yaql.language means legacy yaql
    from muranoclient.common.yaqlexpression_legacy import YaqlExpression


LOG = logging.getLogger(__name__)


# Decorator for cli-args
def arg(*args, **kwargs):
    def _decorator(func):
        # Because of the semantics of decorator composition if we just append
        # to the options list positional options will appear to be backwards.
        func.__dict__.setdefault('arguments', []).insert(0, (args, kwargs))
        return func
    return _decorator


def json_formatter(js):
    return jsonutils.dumps(js, indent=2)


def text_wrap_formatter(d):
    return '\n'.join(textwrap.wrap(d or '', 55))


def pretty_choice_list(l):
    return ', '.join("'%s'" % i for i in l)


def print_list(objs, fields, field_labels, formatters=None, sortby=0):
    if formatters is None:
        formatters = {}
    pt = prettytable.PrettyTable([f for f in field_labels], caching=False)
    pt.align = 'l'

    for o in objs:
        row = []
        for field in fields:
            if field in formatters:
                row.append(formatters[field](o))
            else:
                data = getattr(o, field, None) or ''
                row.append(data)
        pt.add_row(row)

    result = encodeutils.safe_encode(pt.get_string())

    result = result.decode()

    print(result)


def print_dict(d, formatters=None):
    if formatters is None:
        formatters = {}
    pt = prettytable.PrettyTable(['Property', 'Value'], caching=False)
    pt.align = 'l'

    for field in d.keys():
        if field in formatters:
            pt.add_row([field, formatters[field](d[field])])
        else:
            pt.add_row([field, d[field]])

    result = encodeutils.safe_encode(pt.get_string(sortby='Property'))

    result = result.decode()

    print(result)


def find_resource(manager, name_or_id, *args, **kwargs):
    """Helper for the _find_* methods."""
    # first try to get entity as integer id
    try:
        if isinstance(name_or_id, int) or name_or_id.isdigit():
            return manager.get(int(name_or_id), *args, **kwargs)
    except exceptions.NotFound:
        pass

    # now try to get entity as uuid
    try:
        uuid.UUID(str(name_or_id))
        return manager.get(name_or_id, *args, **kwargs)
    except (ValueError, exceptions.NotFound):
        pass

    # finally try to find entity by name
    try:
        return manager.find(name=name_or_id)
    except exceptions.NotFound:
        msg = "No %s with a name or ID of '%s' exists." % \
              (manager.resource_class.__name__.lower(), name_or_id)
        raise exceptions.CommandError(msg)


def string_to_bool(arg):
    return arg.strip().lower() in ('t', 'true', 'yes', '1')


def env(*vars, **kwargs):
    """Search for the first defined of possibly many env vars

    Returns the first environment variable defined in vars, or
    returns the default defined in kwargs.
    """
    for v in vars:
        value = os.environ.get(v, None)
        if value:
            return value
    return kwargs.get('default', '')


def exit(msg=''):
    if msg:
        print(encodeutils.safe_encode(msg), file=sys.stderr)
    sys.exit(1)


def getsockopt(self, *args, **kwargs):
    """Allows us to monkey patch eventlet's GreenSocket

    A function which allows us to monkey patch eventlet's
    GreenSocket, adding a required 'getsockopt' method.
    TODO: (mclaren) we can remove this once the eventlet fix
    (https://bitbucket.org/eventlet/eventlet/commits/609f230)
    lands in mainstream packages.
    NOTE: Already in 0.13, but we can't be sure that all clients
    that use python-muranoclient also use newest eventlet
    """
    return self.fd.getsockopt(*args, **kwargs)


def exception_to_str(exc):
    try:
        error = str(exc)
    except UnicodeError:
        error = ("Caught '%(exception)s' exception." %
                 {"exception": exc.__class__.__name__})
    return encodeutils.safe_encode(error, errors='ignore')


class NoCloseProxy(object):
    """A proxy object, that does nothing on close."""
    def __init__(self, obj):
        self.obj = obj

    def close(self):
        return

    def __getattr__(self, name):
        return getattr(self.obj, name)


class File(object):
    def __init__(self, name, binary=True):
        self.name = name
        self.binary = binary

    def open(self):
        mode = 'rb' if self.binary else 'r'
        if hasattr(self.name, 'read'):
            # NOTE(kzaitsev) We do not want to close a file object
            # passed to File wrapper. The caller should be responsible
            # for closing it
            return NoCloseProxy(self.name)
        else:
            if os.path.isfile(self.name):
                return open(self.name, mode)
            if os.path.isdir(self.name):
                tmp = tempfile.NamedTemporaryFile()
                archive = zipfile.ZipFile(tmp, 'w', zipfile.ZIP_DEFLATED)
                for root, dirs, files in os.walk(self.name):
                    for _file in files:
                        destination = os.path.relpath(
                            os.path.join(root, _file), os.path.join(self.name))
                        archive.write(os.path.join(root, _file), destination)
                tmp.flush()
                return open(tmp.name, mode)
            url = urllib.parse.urlparse(self.name)
            if url.scheme in ('http', 'https'):
                resp = requests.get(self.name, stream=True)
                if not resp.ok:
                    raise ValueError("Got non-ok status({0}) "
                                     "while connecting to {1}".format(
                                         resp.status_code, self.name))
                temp_file = tempfile.NamedTemporaryFile(mode='w+b')
                for chunk in resp.iter_content(1024 * 1024):
                    temp_file.write(chunk)
                temp_file.flush()
                return open(temp_file.name, mode)
            raise ValueError("Can't open {0}".format(self.name))


def to_url(filename, base_url, version='', path='/', extension=''):
    if urllib.parse.urlparse(filename).scheme in ('http', 'https'):
        return filename
    if not base_url:
        raise ValueError("No base_url for repository supplied")
    if '/' in filename or filename in ('.', '..'):
        raise ValueError("Invalid filename path supplied: {0}".format(
            filename))
    version = '.' + version if version else ''
    return urllib.parse.urljoin(base_url,
                                path + filename + version + extension)


class FileWrapperMixin(object):
    def __init__(self, file_wrapper):
        self.file_wrapper = file_wrapper
        try:
            self._file = self.file_wrapper.open()
        except Exception:
            # NOTE(kzaitsev): We need to have _file available at __del__ time.
            self._file = None
            raise

    def file(self):
        self._file.seek(0)
        return self._file

    def close(self):
        if self._file and not self._file.closed:
            self._file.close()

    def save(self, dst, binary=True):
        file_name = self.file_wrapper.name

        if urllib.parse.urlparse(file_name).scheme:
            file_name = file_name.split('/')[-1]

        dst = os.path.join(dst, file_name)

        mode = 'wb' if binary else 'w'
        with open(dst, mode) as dst_file:
            self._file.seek(0)
            shutil.copyfileobj(self._file, dst_file)

    def __del__(self):
        self.close()


class Package(FileWrapperMixin):
    """Represents murano package contents."""

    @staticmethod
    def from_file(file_obj):
        if not isinstance(file_obj, File):
            file_obj = File(file_obj)
        pkg = Package(file_obj)
        errs = pkg.validate()
        if errs:
            raise exceptions.HTTPBadRequest(details=errs)
        return pkg

    @staticmethod
    def fromFile(file_obj):
        warnings.warn("Use from_file function", DeprecationWarning)
        return Package.from_file(file_obj)

    @staticmethod
    def from_location(name, base_url='', version='', url='', path=None):
        """Open file using one of three possible options

        If path is supplied search for name file in the path, otherwise
        if url is supplied - open that url and finally search murano
        repository for the package.
        """
        if path:
            pkg_name = os.path.join(path, name)
            file_name = None
            for f in [pkg_name, pkg_name + '.zip']:
                if os.path.exists(f):
                    file_name = f
            if file_name:
                return Package.from_file(file_name)
            LOG.error("Couldn't find file for package {0}, tried {1}".format(
                name, [pkg_name, pkg_name + '.zip']))
        if url:
            return Package.from_file(url)
        return Package.from_file(to_url(
            name,
            base_url=base_url,
            version=version,
            path='apps/',
            extension='.zip')
        )

    def validate(self):
        m = check_manager.Manager(self._file,
                                  loader=check_pkg_loader.ZipLoader)
        errors = m.validate(
            validators=[check_validators.manifest.ManifestValidator],
            only_errors=True)
        if errors:
            fmt = check_manager.PlainTextFormatter().format
            return 'Invalid Murano package\n{}\n'.format(fmt(errors))

    @property
    def contents(self):
        """Contents of a package."""
        if not hasattr(self, '_contents'):
            try:
                self._file.seek(0)
                self._zip_obj = zipfile.ZipFile(
                    BytesIO(self._file.read()))
            except Exception as e:
                LOG.error("Error {0} occurred,"
                          " while parsing the package".format(e))
                raise
        return self._zip_obj

    @property
    def manifest(self):
        """Parsed manifest file of a package."""
        if not hasattr(self, '_manifest'):
            try:
                self._manifest = yaml.safe_load(
                    self.contents.open('manifest.yaml'))
            except Exception as e:
                LOG.error("Error {0} occurred, while extracting "
                          "manifest from package".format(e))
                raise
        return self._manifest

    def images(self):
        """Returns a list of required image specifications."""
        if 'images.lst' not in self.contents.namelist():
            return []
        try:
            return yaml.safe_load(
                self.contents.open('images.lst')).get('Images', [])
        except Exception:
            return []

    @property
    def resolvers(self):
        if not hasattr(self, '_resolvers'):
            self.classes
        return self._resolvers

    @property
    def classes(self):
        if not hasattr(self, '_classes'):
            self._classes = {}
            self._resolvers = {}
            for class_name, class_file in (
                    self.manifest.get('Classes', {}).items()):
                filename = "Classes/%s" % class_file
                if filename not in self.contents.namelist():
                    continue
                klass_list = yaml.load_all(self.contents.open(filename),
                                           DummyYaqlYamlLoader)
                if not klass_list:
                    raise ValueError('No classes defined in file')
                resolver = None
                for klass in klass_list:
                    ns = klass.get('Namespaces')
                    if ns:
                        resolver = NamespaceResolver(ns)
                    name = klass.get('Name')
                    if name and resolver:
                        name = resolver.resolve_name(name)
                    if name == class_name:
                        self._classes[class_name] = klass
                        self._resolvers[class_name] = resolver
                        break
        return self._classes

    @property
    def ui(self):
        if not hasattr(self, '_ui'):
            if 'UI/ui.yaml' in self.contents.namelist():
                self._ui = self.contents.open('UI/ui.yaml')
            else:
                self._ui = None
        return self._ui

    @property
    def logo(self):
        if not hasattr(self, '_logo'):
            if 'logo.png' in self.contents.namelist():
                self._logo = self.contents.open('logo.png')
            else:
                self._logo = None
        return self._logo

    def _get_package_order(self, packages_graph):
        """Sorts packages according to dependencies between them

        Murano allows cyclic dependencies. It is impossible
        to do topological sort for graph with cycles, so at first
        graph condensation should be built.
        For condensation building Kosaraju's algorithm is used.
        Packages in strongly connected components can be situated
        in random order to each other.
        """
        def topological_sort(graph, start_node):
            order = []
            not_seen = set(graph)

            def dfs(node):
                not_seen.discard(node)
                for dep_node in graph[node]:
                    if dep_node in not_seen:
                        dfs(dep_node)
                order.append(node)

            dfs(start_node)
            return order

        def transpose_graph(graph):
            transposed = collections.defaultdict(list)
            for node, deps in graph.items():
                for dep in deps:
                    transposed[dep].append(node)
            return transposed

        order = topological_sort(packages_graph, self.manifest['FullName'])
        order.reverse()
        transposed = transpose_graph(packages_graph)

        def top_sort_by_components(graph, component_order):
            result = []
            seen = set()

            def dfs(node):
                seen.add(node)
                result.append(node)
                for dep_node in graph[node]:
                    if dep_node not in seen:
                        dfs(dep_node)
            for item in component_order:
                if item not in seen:
                    dfs(item)
            return reversed(result)
        return top_sort_by_components(transposed, order)

    def requirements(self, base_url, path=None, dep_dict=None):
        """Scans Require section of manifests of all the dependencies.

        Returns a dict with FQPNs as keys and respective Package objects
        as values, ordered by topological sort.

        :param base_url: url of packages location
        :param path: local path of packages location
        :param dep_dict: unused. Left for backward compatibility
        """

        unordered_requirements = {}
        requirements_graph = collections.defaultdict(list)
        dep_queue = collections.deque([(self.manifest['FullName'], self)])
        while dep_queue:
            dep_name, dep_file = dep_queue.popleft()
            unordered_requirements[dep_name] = dep_file
            direct_deps = Package._get_direct_deps(dep_file, base_url, path)
            for name, file in direct_deps:
                if name not in unordered_requirements:
                    dep_queue.append((name, file))
            requirements_graph[dep_name] = [dep[0] for dep in direct_deps]

        ordered_reqs_names = self._get_package_order(requirements_graph)
        ordered_reqs_dict = collections.OrderedDict()
        for name in ordered_reqs_names:
            ordered_reqs_dict[name] = unordered_requirements[name]

        return ordered_reqs_dict

    @staticmethod
    def _get_direct_deps(package, base_url, path):
        result = []
        if 'Require' in package.manifest:
            for dep_name, ver in package.manifest['Require'].items():
                try:
                    req_file = Package.from_location(
                        dep_name,
                        version=ver,
                        path=path,
                        base_url=base_url,
                    )
                except Exception as e:
                    LOG.error("Error {0} occurred while parsing package {1}, "
                              "required by {2} package".format(
                                  e, dep_name,
                                  package.manifest['FullName']))
                    continue
                result.append((req_file.manifest['FullName'], req_file))
        return result


def save_image_local(image_spec, base_url, dst):
    dst = os.path.join(dst, image_spec['Name'])

    download_url = to_url(
        image_spec.get("Url", image_spec['Name']),
        base_url=base_url,
        path='images/'
    )

    with open(dst, "w") as image_file:
        response = requests.get(download_url, stream=True)
        total_length = response.headers.get('content-length')

        if total_length is None:
            image_file.write(response.content)
        else:
            dl = 0
            total_length = int(total_length)
            for chunk in response.iter_content(1024 * 1024):
                dl += len(chunk)
                image_file.write(chunk)
                done = int(50 * dl / total_length)
                sys.stdout.write("\r[{0}{1}]".
                                 format('=' * done, ' ' * (50 - done)))
                sys.stdout.flush()
            sys.stdout.write("\n")
            image_file.flush()


def ensure_images(glance_client, image_specs, base_url,
                  local_path=None,
                  is_package_public=False):
    """Ensure that images are available

    Ensure that images from image_specs are available in glance. If not
    attempts: instructs glance to download the images and sets murano-specific
    metadata for it.
    """
    def _image_valid(image, keys):
        for key in keys:
            if key not in image:
                LOG.warning("Image specification invalid: "
                            "No {0} key in image ").format(key)
                return False
        return True

    keys = ['Name', 'DiskFormat', 'ContainerFormat', ]
    installed_images = []
    for image_spec in image_specs:
        if not _image_valid(image_spec, keys):
            continue
        filters = {
            'name': image_spec["Name"],
            'disk_format': image_spec["DiskFormat"],
            'container_format': image_spec["ContainerFormat"],
        }

        images = glance_client.images.list(filters=filters)
        try:
            img = next(images).to_dict()
        except StopIteration:
            img = None

        update_metadata = False
        if img:
            LOG.info("Found desired image {0}, id {1}".format(
                img['name'], img['id']))
            # check for murano meta-data
            if 'murano_image_info' in img.get('properties', {}):
                LOG.info("Image {0} already has murano meta-data".format(
                    image_spec['Name']))
            else:
                update_metadata = True
        else:
            LOG.info("Desired image {0} not found attempting "
                     "to download".format(image_spec['Name']))
            update_metadata = True

            img_file = None
            if local_path:
                img_file = os.path.join(local_path, image_spec['Name'])

            if img_file and not os.path.exists(img_file):
                LOG.error("Image file {0} does not exist."
                          .format(img_file))

            if img_file and os.path.exists(img_file):
                img = glance_client.images.create(
                    name=image_spec['Name'],
                    container_format=image_spec['ContainerFormat'],
                    disk_format=image_spec['DiskFormat'],
                    data=open(img_file, 'rb'),
                )
                img = img.to_dict()
            else:
                download_url = to_url(
                    image_spec.get("Url", image_spec['Name']),
                    base_url=base_url,
                    path='images/',
                )
                LOG.info("Instructing glance to download image {0}".format(
                    image_spec['Name']))
                img = glance_client.images.create(
                    name=image_spec["Name"],
                    container_format=image_spec['ContainerFormat'],
                    disk_format=image_spec['DiskFormat'],
                    copy_from=download_url)
                img = img.to_dict()

            if is_package_public:
                try:
                    glance_client.images.update(img['id'], is_public=True)
                    LOG.debug('Success update for image {0}'.format(img['id']))
                except Exception as e:
                    LOG.exception(_("Error {0} occurred while setting "
                                    "image {1} public").format(e, img['id']))

            installed_images.append(img)

            if update_metadata and 'Meta' in image_spec:
                LOG.info("Updating image {0} metadata".format(
                    image_spec['Name']))
                murano_image_info = jsonutils.dumps(image_spec['Meta'])
                glance_client.images.update(
                    img['id'], properties={'murano_image_info':
                                           murano_image_info})
    return installed_images


class Bundle(FileWrapperMixin):
    """Represents murano bundle contents."""

    @staticmethod
    def from_file(file_obj):
        if not isinstance(file_obj, File):
            file_obj = File(file_obj, binary=False)
        return Bundle(file_obj)

    @staticmethod
    def fromFile(file_obj):
        warnings.warn("Use from_file function", DeprecationWarning)
        return Bundle.from_file(file_obj)

    def package_specs(self):
        """Get a generator yielding package specifications

        Returns a generator yielding package specifications i.e.
        dicts with 'Name' and 'Version' fields
        """
        self._file.seek(0)
        bundle = None
        try:
            # NOTE(kzaitsev) jsonutils throws a type error here
            # see bug 1515231
            bundle = json.load(self._file)
        except ValueError:
            pass
        if bundle is None:
            try:
                bundle = yaml.safe_load(self._file)
            except yaml.error.YAMLError:
                pass

        if bundle is None or 'Packages' not in bundle:
            raise ValueError("Can't parse bundle contents")

        for package in bundle['Packages']:
            if 'Name' not in package:
                continue
            yield package

    def packages(self, base_url='', path=None):
        """Get a generator yielding Package objects

        Returns a generator, yielding Package objects for each package
        found in the bundle.
        """
        for package in self.package_specs():
            try:
                pkg_obj = Package.from_location(
                    package['Name'],
                    version=package.get('Version'),
                    url=package.get('Url'),
                    path=path,
                    base_url=base_url,
                )

            except Exception as e:
                LOG.error("Error {0} occurred while obtaining "
                          "package {1}".format(e, package['Name']))
                continue
            yield pkg_obj


class DummyYaqlYamlLoader(yaml.SafeLoader):
    """Constructor that treats !yaql as string."""
    pass


DummyYaqlYamlLoader.add_constructor(
    u'!yaql', DummyYaqlYamlLoader.yaml_constructors[u'tag:yaml.org,2002:str'])


class YaqlYamlLoader(yaml.SafeLoader):
    pass


# workaround for PyYAML bug: http://pyyaml.org/ticket/221
resolvers = {}
for k, v in yaml.SafeLoader.yaml_implicit_resolvers.items():
    resolvers[k] = v[:]
YaqlYamlLoader.yaml_implicit_resolvers = resolvers


def yaql_constructor(loader, node):
    value = loader.construct_scalar(node)
    return YaqlExpression(value)


YaqlYamlLoader.add_constructor(u'!yaql', yaql_constructor)
YaqlYamlLoader.add_implicit_resolver(u'!yaql', YaqlExpression, None)


def traverse_and_replace(obj,
                         pattern=re.compile(r'^===id(\d+)===$'),
                         replacements=None):
    """Helper function that traverses object model and substitutes ids.

    Recursively checks values of objects found in `obj` against `pattern`,
    and replaces strings that match pattern with uuidutils.generate_uuid().
    Keeps track of any replacements already made, i.e. ===id1=== would
    always be the same, across `obj`. Uses 1st group, found in the `pattern`
    regexp as unique identifier of a replacement
    """
    if replacements is None:
        replacements = collections.defaultdict(
            lambda: uuidutils.generate_uuid(dashed=False))

    def _maybe_replace(obj, key, value):
        """Check and replace value against pattern"""
        if isinstance(value, str):
            m = pattern.search(value)
            if m:
                if m.group(1) not in replacements:
                    replacements[m.group(1)] = uuidutils.generate_uuid(
                        dashed=False)
                obj[key] = replacements[m.group(1)]

    if isinstance(obj, list):
        for key, value in enumerate(obj):
            if isinstance(value, (list, dict)):
                traverse_and_replace(value, pattern, replacements)
            else:
                _maybe_replace(obj, key, value)
    elif isinstance(obj, dict):
        for key, value in obj.items():
            if isinstance(value, (list, dict)):
                traverse_and_replace(value, pattern, replacements)
            else:
                _maybe_replace(obj, key, value)
    else:
        _maybe_replace(obj, key, value)


class NamespaceResolver(object):
    """Copied from main murano repo

    original at murano/dsl/namespace_resolver.py
    """

    def __init__(self, namespaces):
        self._namespaces = namespaces
        self._namespaces[''] = ''

    def resolve_name(self, name, relative=None):
        if name is None:
            raise ValueError()
        if name and name.startswith(':'):
            return name[1:]
        if ':' in name:
            parts = name.split(':')
            if len(parts) != 2 or not parts[1]:
                raise NameError('Incorrectly formatted name ' + name)
            if parts[0] not in self._namespaces:
                raise KeyError('Unknown namespace prefix ' + parts[0])
            return '.'.join((self._namespaces[parts[0]], parts[1]))
        if not relative and '=' in self._namespaces and '.' not in name:
            return '.'.join((self._namespaces['='], name))
        if relative and '.' not in name:
            return '.'.join((relative, name))
        return name



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import zipfile

import yaml


MANIFEST = {'Format': 'MuranoPL/1.0',
            'Type': 'Application',
            'Description': 'MockApp for CLI tests',
            'Author': 'Mirantis, Inc'}


def compose_package(app_name, manifest, package_dir,
                    require=None, archive_dir=None):
    """Composes a murano package

    Composes package `app_name` with `manifest` file as a template for the
    manifest and files from `package_dir`.
    Includes `require` section if any in the manifest file.
    Puts the resulting .zip file into `archive_dir` if present or in the
    `package_dir`.
    """
    with open(manifest, 'w') as f:
        fqn = 'io.murano.apps.' + app_name
        mfest_copy = MANIFEST.copy()
        mfest_copy['FullName'] = fqn
        mfest_copy['Name'] = app_name
        mfest_copy['Classes'] = {fqn: 'mock_muranopl.yaml'}
        if require:
            mfest_copy['Require'] = require
        f.write(yaml.dump(mfest_copy, default_flow_style=False))

    name = app_name + '.zip'

    if not archive_dir:
        archive_dir = os.path.dirname(os.path.abspath(__file__))
    archive_path = os.path.join(archive_dir, name)

    with zipfile.ZipFile(archive_path, 'w') as zip_file:
        for root, dirs, files in os.walk(package_dir):
            for f in files:
                zip_file.write(
                    os.path.join(root, f),
                    arcname=os.path.join(os.path.relpath(root, package_dir), f)
                )

    return archive_path



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import unittest

from muranoclient.tests.functional.cli import \
    murano_test_utils as utils
from muranoclient.tests.functional import muranoclient as murano_client
from oslo_utils.strutils import bool_from_string as str2bool


# TODO(mstolyarenko):  need to remove this raw when
# https://bugs.launchpad.net/python-muranoclient/+bug/1625039 is fixed
backend_name =\
    murano_client.ClientTestBase.get_backend_flag().rstrip().split()[-1]


class SimpleReadOnlyMuranoClientTest(utils.CLIUtilsTestBase):
    """Basic, read-only tests for Murano CLI client.

    Basic smoke test for the Murano CLI commands which do not require
    creating or modifying murano objects.
    """

    def test_category_list(self):
        category = self.get_table_struct('category-list')
        self.assertEqual(['ID', 'Name'], category)

    def test_env_template_list(self):
        templates = self.get_table_struct('env-template-list')
        self.assertEqual(['ID', 'Name', 'Created', 'Updated', 'Is public'],
                         templates)

    def test_environment_list(self):
        environment = self.get_table_struct('environment-list')
        self.assertEqual(['ID', 'Name', 'Status', 'Created', 'Updated'],
                         environment)

    def test_package_list(self):
        packages = self.get_table_struct('package-list')
        self.assertEqual(['ID', 'Name', 'FQN', 'Author', 'Active',
                          'Is Public', 'Type', 'Version'], packages)


class TableStructureMuranoClientTest(utils.CLIUtilsTestBase):
    """Smoke test for the Murano CLI commands

    Smoke test for the Murano CLI commands which checks table
    structure after create or delete category, env-template
    environment and package.
    """

    def test_table_struct_deployment_list(self):
        """Test scenario:

            1) create environment
            2) check table structure
        """
        environment = self.create_murano_object('environment',
                                                'MuranoTestTS-depl-list')
        table_struct = self.get_table_struct('deployment-list',
                                             params=environment['ID'])
        self.assertEqual(['ID', 'State', 'Created', 'Updated', 'Finished'],
                         table_struct)

    def test_table_struct_of_environment_create(self):
        """Test scenario:

            1) create environment
            2) check table structure
        """
        self.create_murano_object('environment', 'MuranoTestTS-env-create')
        table_struct = self.get_table_struct('environment-list')
        self.assertEqual(['ID', 'Name', 'Status', 'Created', 'Updated'],
                         table_struct)

    def test_table_struct_of_environment_delete(self):
        """Test scenario:

            1) create environment
            2) delete environment
            3) check table structure
        """
        environment = self.create_murano_object('environment',
                                                'MuranoTestTS-env-del')
        self.delete_murano_object('environment', environment)
        table_struct = self.get_table_struct('environment-list')
        self.assertEqual(['ID', 'Name', 'Status', 'Created', 'Updated'],
                         table_struct)

    def test_table_struct_of_category_create(self):
        """Test scenario:

            1) create category
            2) check table structure
        """
        self.create_murano_object('category', 'MuranoTestTS-cat-create')
        table_struct = self.get_table_struct('category-list')
        self.assertEqual(['ID', 'Name'], table_struct)

    def test_table_struct_of_category_delete(self):
        """Test scenario:

            1) create category
            2) delete category
            3) check table structure
        """
        category = self.create_murano_object('category',
                                             'MuranoTestTS-cat-create')
        self.delete_murano_object('category', category)
        category = self.get_table_struct('category-list')
        self.assertEqual(['ID', 'Name'], category)

    def test_table_struct_of_env_template_create(self):
        """Test scenario:

            1) create env_template
            2) check table structure
        """
        self.create_murano_object('env-template',
                                  'MuranoTestTS-env-tmp-create')
        table_struct = self.get_table_struct('env-template-list')
        self.assertEqual(['ID', 'Name', 'Created', 'Updated', 'Is public'],
                         table_struct)

    def test_table_struct_of_env_template_delete(self):
        """Test scenario:

            1) create env_template
            2) delete env_template
            3) check table structure
        """
        env_template = self.create_murano_object('env-template',
                                                 'MuranoTestTS-env-tmp-create')
        self.delete_murano_object('env-template', env_template)
        table_struct = self.get_table_struct('env-template-list')
        self.assertEqual(['ID', 'Name', 'Created', 'Updated', 'Is public'],
                         table_struct)


class EnvironmentMuranoSanityClientTest(utils.CLIUtilsTestBase):
    """Sanity tests for testing actions with environment.

    Smoke test for the Murano CLI commands which checks basic actions with
    environment command like create, delete, rename etc.
    """

    def test_environment_create(self):
        """Test scenario:

            1) create environment
            2) check that created environment exist
        """
        environment = self.create_murano_object('environment',
                                                'TestMuranoSanityEnv')
        env_list = self.listing('environment-list')

        # Deleting dates from dictionaries to skip it in assert
        list(map(lambda x: x.pop('Updated', None),
             env_list + [environment]))
        list(map(lambda x: x.pop('Created', None),
             env_list + [environment]))

        self.assertIn(environment, env_list)

    def test_environment_delete(self):
        """Test scenario:

            1) create environment
            2) delete environment
        """
        environment = self.create_murano_object('environment',
                                                'TestMuranoSanityEnv')
        self.delete_murano_object('environment', environment)
        env_list = self.listing('environment-list')

        self.assertNotIn(environment, env_list)

    def test_environment_rename(self):
        """Test scenario:

            1) create environment
            2) rename environment
        """
        environment = self.create_murano_object('environment',
                                                'TestMuranoSanityEnv')

        new_env_name = self.generate_name('TestMuranoSEnv-env-rename')
        rename_params = "{0} {1}".format(environment['Name'], new_env_name)
        new_list = self.listing('environment-rename', params=rename_params)
        renamed_env = self.get_object(new_list, new_env_name)
        self.addCleanup(self.delete_murano_object, 'environment', renamed_env)
        new_env_list = self.listing('environment-list')

        # Deleting dates from dictionaries to skip it in assert
        list(map(lambda x: x.pop('Updated', None),
             new_env_list + [environment] + [renamed_env]))
        list(map(lambda x: x.pop('Created', None),
             new_env_list + [environment] + [renamed_env]))

        self.assertIn(renamed_env, new_env_list)
        self.assertNotIn(environment, new_env_list)

    def test_table_struct_env_show(self):
        """Test scenario:

            1) create environment
            2) check structure of env_show object
        """
        environment = self.create_murano_object('environment',
                                                'TestMuranoSanityEnv')
        env_show = self.listing('environment-show', params=environment['Name'])
        # Check structure of env_show object
        self.assertEqual(['acquired_by', 'created', 'description_text', 'id',
                          'name', 'services', 'status', 'tenant_id',
                          'updated', 'version'],
                         list(map(lambda x: x['Property'], env_show)))

    def test_environment_show(self):
        """Test scenario:

            1) create environment
            2) check that env_name, ID, updated and created values
               exist in env_show object
        """
        environment = self.create_murano_object('environment',
                                                'TestMuranoSanityEnv')
        env_show = self.listing('environment-show', params=environment['Name'])

        self.assertIn(environment['Created'],
                      list(map(lambda x: x['Value'], env_show)))
        self.assertIn(environment['Updated'],
                      list(map(lambda x: x['Value'], env_show)))
        self.assertIn(environment['Name'],
                      list(map(lambda x: x['Value'], env_show)))
        self.assertIn(environment['ID'],
                      list(map(lambda x: x['Value'], env_show)))

    def test_environment_delete_by_id(self):
        """Test scenario:

            1) create environment
            2) delete environment by environment ID
        """
        env_name = self.generate_name('TestMuranoSanityEnv')
        environment = self.create_murano_object('environment', env_name)
        result = self.murano('environment-delete', params=environment['ID'],
                             fail_ok=False)
        self.assertNotIn(environment['Name'], result)
        env_list = self.listing('environment-list')
        self.assertNotIn(environment, env_list)

    def test_environment_model_show(self):
        """Test scenario:

            1) create environment
            2) check that the result of environment-model-show is a valid
               non-empty json
        """
        env_name = self.generate_name('TestMuranoSanityEnv')
        environment = self.create_murano_object('environment', env_name)
        model = self.murano('environment-model-show', params=environment['ID'])
        result = json.loads(model)
        self.assertEqual(4, len(result))


class CategoryMuranoSanityClientTest(utils.CLIUtilsTestBase):
    """Sanity tests for testing actions with Category.

    Smoke test for the Murano CLI commands which checks basic actions with
    category command like create, delete etc.
    """

    def test_category_create(self):
        """Test scenario:

            1) create category
            2) check that created category exist
        """
        category = self.create_murano_object('category',
                                             'TestMuranoSanityCategory')
        category_list = self.listing('category-list')

        self.assertIn(category, category_list)

    def test_category_delete(self):
        """Test scenario:

            1) create category
            2) delete category
            3) check that category has been deleted successfully
        """
        category = self.create_murano_object('category',
                                             'TestMuranoSanityCategory')
        self.delete_murano_object('category', category)
        category_list = self.listing('category-list')

        self.assertNotIn(category, category_list)

    def test_table_struct_category_show(self):
        """Test scenario:

            1) create category
            2) check table structure of category-show object
        """
        category = self.create_murano_object('category',
                                             'TestMuranoSanityCategory')
        category_show = self.listing('category-show', params=category['ID'])

        self.assertEqual(['id', 'name', 'packages'],
                         list(map(lambda x: x['Property'], category_show)))

    def test_category_show(self):
        """Test scenario:

            1) create category
            2) check that category values exist in category_show object
        """
        category = self.create_murano_object('category',
                                             'TestMuranoSanityCategory')
        category_show = self.listing('category-show', params=category['ID'])

        self.assertIn(category['ID'],
                      list(map(lambda x: x['Value'], category_show)))
        self.assertIn(category['Name'],
                      list(map(lambda x: x['Value'], category_show)))

    def test_non_existing_category_delete(self):
        """Test scenario:

            1) try to call category-delete for non existing category
            2) check that error message contains user friendly substring
        """
        result = self.murano('category-delete', params='non-existing',
                             fail_ok=True)
        self.assertIn("Failed to delete 'non-existing'; category not found",
                      result)

    def test_non_existing_category_show(self):
        """Test scenario:

            1) try to call category-show for non existing category
            2) check that error message contains user friendly substring
        """
        result = self.murano('category-show', params='non-existing',
                             fail_ok=True)
        self.assertIn("Category id 'non-existing' not found", str(result))

    def test_category_create_with_long_name(self):
        """Test scenario:

            1) try to create category with long name (>80)
            2) check that error message contains user friendly substring
        """
        result = self.murano('category-create', params='name' * 21,
                             fail_ok=True)
        self.assertIn(
            "Category name should be 80 characters maximum",
            result)


class EnvTemplateMuranoSanityClientTest(utils.CLIUtilsTestBase):
    """Sanity tests for testing actions with Environment template.

    Smoke test for the Murano CLI commands which checks basic actions with
    env-temlate command like create, delete etc.
    """
    def test_environment_template_create(self):
        """Test scenario:

            1) create environment template
            2) check that created environment template exist
        """
        env_template = self.create_murano_object('env-template',
                                                 'TestMuranoSanityEnvTemp')
        env_template_list = self.listing('env-template-list')

        # Deleting dates from dictionaries to skip it in assert
        list(map(lambda x: x.pop('Updated', None),
             env_template_list + [env_template]))
        list(map(lambda x: x.pop('Created', None),
             env_template_list + [env_template]))

        self.assertIn(env_template, env_template_list)

    def test_environment_template_delete(self):
        """Test scenario:

            1) create environment template
            2) delete environment template
            3) check that deleted environment template doesn't exist
        """
        env_template = self.create_murano_object('env-template',
                                                 'TestMuranoSanityEnvTemp')
        env_template_list = self.delete_murano_object('env-template',
                                                      env_template)

        self.assertNotIn(env_template, env_template_list)

    def test_table_struct_env_template_show(self):
        """Test scenario:

            1) create environment template
            2) check table structure of env-template-show object
        """
        env_template = self.create_murano_object('env-template',
                                                 'TestMuranoSanityEnvTemp')
        env_template_show = self.listing('env-template-show',
                                         params=env_template['ID'])
        tested_env_template = list(
            map(lambda x: x['Property'], env_template_show))

        self.assertIn('created', tested_env_template)
        self.assertIn('id', tested_env_template)
        self.assertIn('name', tested_env_template)
        self.assertIn('services', tested_env_template)
        self.assertIn('tenant_id', tested_env_template)
        self.assertIn('updated', tested_env_template)
        self.assertIn('version', tested_env_template)

    def test_env_template_show(self):
        """Test scenario:

            1) create environment template
            2) check that environment template values exist in
            env-template-show object
        """
        env_template = self.create_murano_object('env-template',
                                                 'TestMuranoSanityEnvTemp')
        env_template_show = self.listing('env-template-show',
                                         params=env_template['ID'])
        tested_env = list(map(lambda x: x['Value'], env_template_show))

        self.assertIn(env_template['ID'], tested_env)
        self.assertIn(env_template['Name'], tested_env)

    def test_env_template_create_environment(self):
        """Test scenario:

            1) create environment template
            2) create environment from template
        """
        env_template = self.create_murano_object('env-template',
                                                 'TestMuranoSanityEnvTemp')
        new_env_name = self.generate_name('EnvFromTemp')
        params = "{0} {1}".format(env_template['ID'], new_env_name)
        env_created = self.listing('env-template-create-env', params=params)
        tested_env_created = list(map(lambda x: x['Property'], env_created))

        self.assertIn('environment_id', tested_env_created)
        self.assertIn('session_id', tested_env_created)

    def test_env_template_clone(self):
        """Test scenario:

            1) create environment template
            2) clone template
            3) check that create environment template has the new name
            4) delete new template
        """

        env_template = self.create_murano_object_parameter(
            'env-template', 'TestMuranoSanityEnvTemp', '--is-public')
        new_template = self.generate_name('TestMuranoSanityEnvTemp')

        params = "{0} {1}".format(env_template['ID'], new_template)
        template_created = self.listing('env-template-clone', params=params)
        tp_list = list(
            map(lambda x: ({x['Property']: x['Value']}), template_created))
        result_name = list(filter(lambda x: x.get('name'), tp_list))[0]['name']
        result_id = list(filter(lambda x: x.get('id'), tp_list))[0]['id']
        self.listing('env-template-delete', params=result_id)

        self.assertIn(result_name, new_template)


class PackageMuranoSanityClientTest(utils.CLIUtilsTestPackagesBase):
    """Sanity tests for testing actions with Packages.

    Smoke tests for the Murano CLI commands which check basic actions with
    packages like import, create, delete etc.
    """

    def test_package_import_by_url(self):
        """Test scenario:

            1) import package by url
            2) check that package exists
        """
        try:
            self.run_server()
            package = self.import_package(
                self.app_name,
                'http://localhost:8089/apps/{0}.zip'.format(self.app_name)
            )
        finally:
            self.stop_server()
        package_list = self.listing('package-list')

        self.assertIn(package, package_list)

    def test_package_import_by_path(self):
        """Test scenario:

            1) import package by path
            2) check that package exists
        """
        package = self.import_package(
            self.app_name,
            self.dummy_app_path
        )
        package_list = self.listing('package-list')

        self.assertIn(package, package_list)

    def test_package_is_public(self):
        """Test scenario:

            1) import package
            2) check that package is public
        """
        package = self.import_package(
            self.app_name,
            self.dummy_app_path,
            '--is-public')

        package_show = self.listing('package-show', params=package['ID'])
        package_show = {item['Property']: item['Value']
                        for item in package_show}

        self.assertEqual(package['Is Public'], 'True')
        self.assertEqual(
            str2bool(package['Is Public']),
            str2bool(package_show['is_public']))

    def test_package_delete(self):
        """Test scenario:

            1) import package
            2) delete package
            3) check that package has been deleted
        """

        package = self.import_package(
            self.app_name,
            self.dummy_app_path
        )
        package_list = self.delete_murano_object('package', package)

        self.assertNotIn(package, package_list)

    def test_package_show(self):
        """Test scenario:

            1) import package
            2) check that package values exist in
            return by package-show object
        """

        package = self.import_package(
            self.app_name,
            self.dummy_app_path
        )
        package_show = self.listing('package-show', params=package['ID'])
        package_show = {item['Property']: item['Value']
                        for item in package_show}

        self.assertEqual(
            str2bool(package['Active']),
            str2bool(package_show['enabled']))
        self.assertEqual(
            package['FQN'],
            package_show['fully_qualified_name'])
        self.assertEqual(
            package['ID'],
            package_show['id'])
        self.assertEqual(
            str2bool(package['Is Public']),
            str2bool(package_show['is_public']))
        self.assertEqual(
            package['Name'],
            package_show['name'])
        self.assertEqual(
            package['Type'],
            package_show['type'])

    def test_package_import_update(self):
        """Test scenario:

            1) import package
            2) import new_package using option 'u' - update
            3) check that package has been updated
        """
        package = self.import_package(
            self.app_name,
            self.dummy_app_path
        )
        upd_package = self.import_package(
            self.app_name,
            self.dummy_app_path,
            '--exists-action', 'u'
        )
        self.assertEqual(package['Name'], upd_package['Name'])
        self.assertNotEqual(package['ID'], upd_package['ID'])

    def test_package_import_skip(self):
        """Test scenario:

            1) import package using option 's' - skip for existing package
            2) try to import the same package using option 's' - skip
            3) check that package hasn't been updated
        """
        package = self.import_package(
            self.app_name,
            self.dummy_app_path,
            '--exists-action', 's'
        )
        updated_package = self.import_package(
            self.app_name,
            self.dummy_app_path,
            '--exists-action', 's'
        )
        package_list = self.listing("package-list")

        self.assertIn(package, package_list)
        self.assertIsNone(updated_package)

    def test_package_import_abort(self):
        """Test scenario:

            1) import package
            2) import new_package using option 'a' - skip
            3) check that package hasn't been updated
        """
        package = self.import_package(
            self.app_name,
            self.dummy_app_path
        )
        package_list = self.listing('package-list')

        self.assertIn(package, package_list)

        package = self.import_package(
            self.app_name,
            self.dummy_app_path,
            '--exists-action', 'a'
        )
        package_list = self.listing('package-list')
        self.assertNotIn(package, package_list)


class DeployMuranoEnvironmentTest(utils.CLIUtilsTestPackagesBase):
    """Test for testing Murano environment deployment.

    Test for the Murano CLI commands which checks addition of app
    to the environment, session creation and deployment of
    environment.
    """
    # TODO(mstolyarenko):  need to unskip this test when
    # https://bugs.launchpad.net/python-muranoclient/+bug/1625039 is fixed
    @unittest.skipIf(backend_name == 'glare',
                     "This test fails when GLARE is used as packages "
                     "service. To be fixed as part of #1625039")
    def test_environment_deployment(self):
        """Test scenario:

            1) import package
            2) create environment
            3) create session for created environment
            4) add application to the environment
            5) send environment to deploy
            6) check that deployment was successful
        """
        self.import_package(
            self.app_name,
            self.dummy_app_path
        )

        env_id = self.create_murano_object('environment',
                                           'TestMuranoDeployEnv')['ID']
        obj_model = {
            'op': 'add',
            'path': '/-',
            'value': {
                '?': {
                    'type': 'io.murano.apps.{0}'.format(self.app_name),
                    'id': '{0}'.format(self.generate_uuid()),
                }
            }
        }
        self.deploy_environment(env_id, obj_model)
        deployments = self.listing('deployment-list', params=env_id)

        self.assertEqual('success', deployments[0]['State'])
        self.assertEqual(1, len(deployments))

    # TODO(mstolyarenko):  need to unskip this test when
    # https://bugs.launchpad.net/python-muranoclient/+bug/1625039 is fixed
    @unittest.skipIf(backend_name == 'glare',
                     "This test fails when GLARE is used as packages "
                     "service. To be fixed as part of #1625039")
    def test_add_component_to_deployed_env(self):
        """Test scenario:

            1) import package
            2) create environment
            3) create session for created environment
            4) add application to the environment
            5) send environment to deploy
            6) check that deployment was successful
            7) add application to environment
            8) deploy environment again
        """
        self.import_package(
            self.app_name,
            self.dummy_app_path
        )

        env_id = self.create_murano_object('environment',
                                           'TestMuranoDeployEnv')['ID']
        obj_model = {
            'op': 'add',
            'path': '/-',
            'value': {
                '?': {
                    'type': 'io.murano.apps.{0}'.format(self.app_name),
                    'id': '',
                }
            }
        }
        obj_model['value']['?']['id'] = self.generate_uuid()
        self.deploy_environment(env_id, obj_model)

        deployments = self.listing('deployment-list', params=env_id)
        self.assertEqual('success', deployments[0]['State'])
        self.assertEqual(1, len(deployments))

        obj_model['value']['?']['id'] = self.generate_uuid()
        self.deploy_environment(env_id, obj_model)

        deployments = self.listing('deployment-list', params=env_id)
        self.assertEqual('success', deployments[1]['State'])
        self.assertEqual(2, len(deployments))

    # TODO(mstolyarenko):  need to unskip this test when
    # https://bugs.launchpad.net/python-muranoclient/+bug/1625039 is fixed
    @unittest.skipIf(backend_name == 'glare',
                     "This test fails when GLARE is used as packages "
                     "service. To be fixed as part of #1625039")
    def test_delete_component_from_deployed_env(self):
        """Test scenario:

            1) import package
            2) create environment
            3) create session for created environment
            4) add application to the environment
            5) send environment to deploy
            6) check that deployment was successful
            7) delete application from environment
            8) deploy environment again
        """
        self.import_package(
            self.app_name,
            self.dummy_app_path
        )

        env_id = self.create_murano_object('environment',
                                           'TestMuranoDeployEnv')['ID']

        obj_model = {
            'op': 'add',
            'path': '/-',
            'value': {
                '?': {
                    'type': 'io.murano.apps.{0}'.format(self.app_name),
                    'id': '{0}'.format(self.generate_uuid()),
                }
            }
        }
        self.deploy_environment(env_id, obj_model)

        obj_model = {
            'op': 'remove',
            'path': '/0'
        }
        self.deploy_environment(env_id, obj_model)

        deployments = self.listing('deployment-list', params=env_id)
        self.assertEqual('success', deployments[1]['State'])
        self.assertEqual(2, len(deployments))


class BundleMuranoSanityClientTest(utils.CLIUtilsTestPackagesBase):
    """Sanity tests for testing actions with bundle.

    Tests for the Murano CLI commands which check basic actions with
    bundles.
    """

    def test_bundle_import_without_bundle_name(self):
        """Test scenario:

            1) Execute murano bundle-import command without bundle name
            2) check that error message contains user friendly substring
        """
        result = self.murano('bundle-import', params='',
                             fail_ok=True)
        self.assertIn("murano bundle-import: error: the following "
                      "arguments are required", result)

    @unittest.skip("Skip due to apps.openstack.org website is retired.")
    def test_bundle_import_with_non_existing_package_name(self):
        """Test scenario:

            1) Execute murano bundle-import command with non-existing packages
            name inside
            2) check that error message contains user friendly substring
        """
        result = self.murano(
            'bundle-import',
            params=self.prepare_bundle_with_non_existed_package(),
            fail_ok=False)
        self.assertIn("Couldn't find file for package", result)
        self.assertIn("Error Got non-ok status(404) while connecting", result)

    @unittest.skip("Skip due to apps.openstack.org website is retired.")
    def test_bundle_import_with_non_existing_name(self):
        """Test scenario:

            1) Execute murano bundle-import command with non-existing bundle
            name
            2) check that error message contains user friendly substring
        """
        result = self.murano('bundle-import', params=self.app_name,
                             fail_ok=True)
        self.assertIn("Bundle file '{}' does not exist".format(self.app_name),
                      result)
        self.assertIn("reason: Got non-ok status(404) while connecting to",
                      result)

    def test_bundle_import_with_invalid_file_format(self):
        """Test scenario:

            1) Execute murano bundle-import command with invalid bundle file
            format
            2) check that error message contains user friendly substring
        """
        try:
            self.murano(
                'bundle-import',
                params=self.prepare_bundle_with_invalid_format(),
                fail_ok=False)
        except utils.exceptions.CommandFailed as exception:
            self.assertIn("Can't parse bundle contents", str(exception))


class StaticActionMuranoClientTest(utils.CLIUtilsTestPackagesBase):
    """Tests for testing static actions execution.

    Tests for the Murano CLI commands which check the result of sample
    static action execution.
    """

    def test_static_action_call(self):
        """Test scenario:

                1) import package
                2) call static action of the class in that package
                3) check the result of action
        """
        package = self.import_package(
            self.app_name,
            self.dummy_app_path
        )
        result = self.murano(
            'static-action-call', params='{0} staticAction --package-name {1} '
            '--arguments myName=John myAge=28'.format(package['FQN'],
                                                      package['FQN']))
        expected = "Waiting for result...\nStatic action result: Hello, " \
                   "John. In 5 years you will be 33 years old.\n"
        self.assertEqual(expected, result)



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections

from glanceclient import exc as glance_exc
import yaml

from muranoclient.common import exceptions as exc
from muranoclient.common import utils
from muranoclient.i18n import _


def rewrap_http_exceptions(func):
    def inner(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except glance_exc.HTTPException as e:
            raise exc.from_code(e.code)
    return inner


class ArtifactRepo(object):
    def __init__(self, client, tenant=None):
        self.tenant = tenant
        self.client = client

    def create(self, fqn, data, **kwargs):

        package = utils.Package.from_file(data)
        manifest = package.manifest
        package_draft = {
            'name': manifest.get('FullName', fqn),
            'version': manifest.get('Version', '0.0.0'),
            'description': manifest.get('Description'),
            'display_name': manifest.get('Name', fqn),
            'type': manifest.get('Type', 'Application'),
            'author': manifest.get('Author'),
            'tags': manifest.get('Tags', []),
            'class_definitions': package.classes.keys()
        }
        for k, v in kwargs.items():
            package_draft[k] = v

        inherits = self._get_local_inheritance(package.classes,
                                               package.resolvers)

        # check for global inheritance
        ancestor_queue = collections.deque(inherits.keys())
        while ancestor_queue:
            ancestor_name = ancestor_queue.popleft()
            child_classes = inherits[ancestor_name]

            ancestors = self.list(class_definitions=ancestor_name)
            for ancestor in ancestors:
                # check if ancestor inherits anything
                ancestor_inherits = \
                    ancestor.type_specific_properties.get('inherits', {})
                for name, value in ancestor_inherits.items():
                    # check if this is the class we actually inherit
                    if ancestor_name in value:
                        ancestor_queue.append(name)
                        inherits[name] = child_classes

        package_draft['inherits'] = inherits

        keywords = self._keywords_from_display_name(
            package_draft['display_name'])
        keywords.extend(package_draft['tags'])
        package_draft['keywords'] = keywords

        # NOTE(ativelkov): this is very racy, but until we have a chance to
        # enforce uniqueness right in glance this is the only way to do it
        visibility = package_draft.get('visibility', 'private')
        if visibility == 'public':
            filters = {}
        else:
            filters = {'owner': self.tenant}
        existing = self.list(name=package_draft['name'],
                             version=package_draft['version'], **filters)
        try:
            next(existing)
            raise exc.HTTPConflict("Package already exists")
        except StopIteration:
            pass

        res = self.client.artifacts.create(**package_draft)
        app_id = res.id
        self.client.artifacts.upload_blob(app_id, 'archive', package.file())
        if package.logo is not None:
            self.client.artifacts.upload_blob(app_id, 'logo', package.logo)
        if package.ui is not None:
            self.client.artifacts.upload_blob(app_id, 'ui_definition',
                                              package.ui)
        package.file().close()
        self.client.artifacts.active(app_id)

        return self.client.artifacts.get(app_id)

    @staticmethod
    def _get_local_inheritance(classes, resolvers):
        result = {}
        for class_name, klass in classes.items():
            if 'Extends' not in klass:
                continue
            ns = klass.get('Namespaces')
            if ns:
                resolver = utils.NamespaceResolver(ns)
            else:
                resolver = resolvers.get(class_name)

            if isinstance(klass['Extends'], list):
                bases = klass['Extends']
            else:
                bases = [klass['Extends']]
            for base_class in bases:
                if resolver:
                    base_fqn = resolver.resolve_name(base_class)
                else:
                    base_fqn = base_class
                result.setdefault(base_fqn, []).append(class_name)
        return result

    @staticmethod
    def _keywords_from_display_name(display_name):
        return display_name.split()[:10]

    def list(self,
             sort_field='name',
             sort_dir='asc',
             type=None,
             tags=None,
             limit=None,
             page_size=None,
             **filters):
        sort = "%s:%s" % (sort_field, sort_dir)
        if type is not None:
            filters['type'] = type
        if tags is not None:
            filters['tag'] = tags
        return self.client.artifacts.list(sort=sort,
                                          limit=limit,
                                          page_size=page_size,
                                          filters=filters)

    def get(self, app_id):
        return self.client.artifacts.get(app_id)

    def delete(self, app_id):
        return self.client.artifacts.delete(app_id)

    def update(self, app_id, props_to_remove=None, **new_props):
        new_keywords = []
        new_name = new_props.get('display_name')
        new_tags = new_props.get('tags')
        if new_name:
            new_keywords.extend(self._keywords_from_display_name(new_name))
        if new_tags:
            new_keywords.extend(new_tags)
        if new_keywords:
            new_props['keywords'] = new_keywords
        visibility = new_props.get('visibility')
        if visibility == 'public':
            package = self.client.artifacts.get(app_id)
            # NOTE(ativelkov): this is very racy, but until we have a chance to
            # enforce uniqueness right in glance this is the only way to do it
            existing = self.list(name=package.name,
                                 version=package.version,
                                 visibility='public')
            try:
                while True:
                    package = next(existing)
                    if package.id == app_id:
                        continue
                    else:
                        raise exc.HTTPConflict("Package already exists")
            except StopIteration:
                pass
        return self.client.artifacts.update(app_id,
                                            remove_props=props_to_remove,
                                            **new_props)

    def toggle_active(self, app_id):
        old_val = self.get(app_id).type_specific_properties['enabled']
        return self.update(app_id, enabled=(not old_val))

    def toggle_public(self, app_id):
        visibility = self.get(app_id).visibility
        if visibility == 'public':
            return self.update(app_id, visibility='private')
        else:
            return self.update(app_id, visibility='public')

    def download(self, app_id):
        return self.client.artifacts.download_blob(app_id, 'archive')

    def get_ui(self, app_id, loader_cls=None):
        ui_stream = "".join(
            self.client.artifacts.download_blob(app_id, 'ui_definition'))
        if loader_cls is None:
            loader_cls = yaml.SafeLoader
        return yaml.load(ui_stream, loader_cls)

    def get_logo(self, app_id):
        return self.client.artifacts.download_blob(app_id, 'logo')


class PackageManagerAdapter(object):
    def __init__(self, legacy, glare):
        self.legacy = legacy
        self.glare = glare

    def categories(self):
        return self.legacy.categories()

    @rewrap_http_exceptions
    def create(self, data, files):
        is_public = data.pop('is_public', None)
        if is_public is not None:
            data['visibility'] = 'public' if is_public else 'private'
        fqn = list(files.keys())[0]
        pkg = self.glare.create(fqn, files[fqn], **data)
        return PackageWrapper(pkg)

    @rewrap_http_exceptions
    def filter(self, **kwargs):
        kwargs.pop('catalog', None)  # NOTE(ativelkov): Glare ignores 'catalog'
        include_disabled = kwargs.pop('include_disabled', False)
        order_by = kwargs.pop('order_by', None)
        search = kwargs.pop('search', None)
        category = kwargs.pop('category', None)
        fqn = kwargs.pop('fqn', None)
        class_name = kwargs.pop('class_name', None)
        name = kwargs.pop('name', None)
        if category:
            kwargs['categories'] = category
        if search:
            kwargs['keywords'] = search
        if order_by:
            kwargs['sort_field'] = order_by
        if not include_disabled:
            kwargs['enabled'] = True
        if fqn:
            kwargs['name'] = fqn
        if class_name:
            kwargs['class_definitions'] = class_name
        if name:
            kwargs['display_name'] = name
        # if 'owned' is used there should be a filter with 'owner' parameter
        if kwargs.pop('owned', None):
            kwargs['owner'] = self.glare.tenant

        for pkg in self.glare.list(**kwargs):
            yield PackageWrapper(pkg)

    @rewrap_http_exceptions
    def list(self, include_disabled=False):
        return self.filter(include_disabled=include_disabled)

    @rewrap_http_exceptions
    def delete(self, app_id):
        return self.glare.delete(app_id)

    @rewrap_http_exceptions
    def get(self, app_id):
        return PackageWrapper(self.glare.get(app_id))

    @rewrap_http_exceptions
    def update(self, app_id, body, operation='replace'):
        is_public = body.pop('is_public', None)
        name = body.pop('name', None)
        if is_public is not None:
            body['visibility'] = 'public' if is_public else 'private'
        if name is not None:
            body['display_name'] = name

        if operation == 'replace':
            return PackageWrapper(self.glare.update(app_id, None, **body))

    @rewrap_http_exceptions
    def toggle_active(self, app_id):
        return self.glare.toggle_active(app_id)

    @rewrap_http_exceptions
    def toggle_public(self, app_id):
        return self.glare.toggle_public(app_id)

    @rewrap_http_exceptions
    def download(self, app_id):
        return "".join(self.glare.download(app_id))

    @rewrap_http_exceptions
    def get_logo(self, app_id):
        return "".join(self.glare.get_logo(app_id))

    @rewrap_http_exceptions
    def get_ui(self, app_id, loader_cls=None):
        return self.legacy.get_ui(app_id, loader_cls)


class PackageWrapper(object):
    def __init__(self, item):
        self._item = item

    @property
    def updated(self):
        return self._item.updated_at

    @property
    def created(self):
        return self._item.created_at

    @property
    def is_public(self):
        return self._item.visibility == 'public'

    @property
    def name(self):
        return self._item.type_specific_properties['display_name']

    @property
    def fully_qualified_name(self):
        return self._item.name

    @property
    def owner_id(self):
        return self._item.owner

    def __getstate__(self):
        return {"item": self._item}

    def __setstate__(self, state):
        self._item = state['item']

    def __getattr__(self, name):
        if name in self._item.type_specific_properties:
            return self._item.type_specific_properties.get(name)
        else:
            return getattr(self._item, name)

    def to_dict(self):
        keys = ('author', 'categories', 'class_definitions', 'created',
                'description', 'enabled', 'fully_qualified_name', 'id',
                'is_public', 'name', 'owner_id', 'tags', 'type',
                'updated')
        missing_keys = [key for key in keys if not hasattr(self, key)]
        if missing_keys:
            raise KeyError(_("Some attributes are missing in "
                             "%(pkg_name)s: %(attrs)s.") %
                           {'pkg_name': self.name,
                            'attrs': ", ".join(missing_keys)})
        return {key: getattr(self, key) for key in keys}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Application-catalog v1 stack action implementation"""

import json
import sys
import urllib

import jsonpatch
from osc_lib.command import command
from osc_lib import utils
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from muranoclient.apiclient import exceptions
from muranoclient.common import utils as murano_utils


LOG = logging.getLogger(__name__)


class ListEnvironments(command.Lister):
    """Lists environments"""

    def get_parser(self, prog_name):
        parser = super(ListEnvironments, self).get_parser(prog_name)
        parser.add_argument(
            '--all-tenants',
            action='store_true',
            default=False,
            help='List environments from all tenants (admin only).',
        )
        parser.add_argument(
            '--tenant',
            metavar='<TENANT_ID>',
            default=None,
            help='Allows to list environments for a given tenant (admin only).'
        )

        return parser

    def take_action(self, parsed_args):
        LOG.debug("take_action({0})".format(parsed_args))
        client = self.app.client_manager.application_catalog
        data = client.environments.list(
            parsed_args.all_tenants, parsed_args.tenant)

        columns = ('id', 'name', 'status', 'created', 'updated')
        column_headers = [c.capitalize() for c in columns]
        return (
            column_headers,
            list(utils.get_item_properties(
                s,
                columns,
            ) for s in data)
        )


class ShowEnvironment(command.ShowOne):
    """Display environment details"""

    def get_parser(self, prog_name):
        parser = super(ShowEnvironment, self).get_parser(prog_name)
        parser.add_argument(
            "id",
            metavar="<NAME or ID>",
            help=("Name or ID of the environment to display"),
        )
        parser.add_argument(
            "--only-apps",
            action='store_true',
            default=False,
            help="Only print apps of the environment (useful for automation).",
        )
        parser.add_argument(
            "--session-id",
            metavar="<SESSION_ID>",
            default='',
            help="Id of a config session.",
        )

        return parser

    def take_action(self, parsed_args):
        LOG.debug("take_action({0})".format(parsed_args))
        client = self.app.client_manager.application_catalog

        environment = utils.find_resource(client.environments,
                                          parsed_args.id)
        data = client.environments.get(environment.id,
                                       parsed_args.session_id).to_dict()

        data['services'] = jsonutils.dumps(data['services'], indent=2)

        if getattr(parsed_args, 'only_apps', False):
            return(['services'], [data['services']])
        else:
            return self.dict2columns(data)


class RenameEnvironment(command.Lister):
    """Rename an environment."""

    def get_parser(self, prog_name):
        parser = super(RenameEnvironment, self).get_parser(prog_name)
        parser.add_argument(
            'id',
            metavar="<NAME or ID>",
            help="Environment ID or name.",
        )
        parser.add_argument(
            'name',
            metavar="<ENVIRONMENT_NAME>",
            help="A name to which the environment will be renamed.",
        )

        return parser

    def take_action(self, parsed_args):
        LOG.debug("take_action({0})".format(parsed_args))
        client = self.app.client_manager.application_catalog
        environment = utils.find_resource(client.environments,
                                          parsed_args.id)
        data = client.environments.update(environment.id,
                                          parsed_args.name)

        columns = ('id', 'name', 'status', 'created', 'updated')
        column_headers = [c.capitalize() for c in columns]

        return (
            column_headers,
            [utils.get_item_properties(
                data,
                columns,
            )]
        )


class EnvironmentSessionCreate(command.ShowOne):
    """Creates a new configuration session for environment ID."""

    def get_parser(self, prog_name):
        parser = super(EnvironmentSessionCreate, self).get_parser(prog_name)
        parser.add_argument(
            'id',
            metavar="<ID>",
            help="ID of Environment to add session to.",
        )

        return parser

    def take_action(self, parsed_args):
        LOG.debug("take_action({0})".format(parsed_args))
        client = self.app.client_manager.application_catalog

        environment_id = parsed_args.id
        session_id = client.sessions.configure(environment_id).id
        sessionid = murano_utils.text_wrap_formatter(session_id)

        return (['id'], [sessionid])


class EnvironmentCreate(command.Lister):
    """Create an environment."""

    def get_parser(self, prog_name):
        parser = super(EnvironmentCreate, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar="<ENVIRONMENT_ID>",
            help="Environment name.",
        )
        parser.add_argument(
            '--region',
            metavar="<REGION_NAME>",
            help="Name of the target OpenStack region.",
        )
        parser.add_argument(
            '--join-subnet-id',
            metavar="<SUBNET_ID>",
            help="Subnetwork id to join.",
        )
        parser.add_argument(
            '--join-net-id',
            metavar="<NET_ID>",
            help="Network id to join.",
        )

        return parser

    def take_action(self, parsed_args):
        LOG.debug("take_action({0})".format(parsed_args))
        client = self.app.client_manager.application_catalog

        body = {"name": parsed_args.name, "region": parsed_args.region}
        if parsed_args.join_net_id or parsed_args.join_subnet_id:
            res = {
                'defaultNetworks': {
                    'environment': {
                        '?': {
                            'id': uuidutils.generate_uuid(dashed=False),
                            'type':
                            'io.murano.resources.ExistingNeutronNetwork'
                        },
                    },
                    'flat': None
                }
            }
            if parsed_args.join_net_id:
                res['defaultNetworks']['environment']['internalNetworkName'] \
                    = parsed_args.join_net_id
            if parsed_args.join_subnet_id:
                res['defaultNetworks']['environment']['internalSubnetworkName'
                                                      ] = \
                    parsed_args.join_subnet_id

            body.update(res)

        data = client.environments.create(body)

        columns = ('id', 'name', 'status', 'created', 'updated')
        column_headers = [c.capitalize() for c in columns]

        return (
            column_headers,
            [utils.get_item_properties(
                data,
                columns,
            )]
        )


class EnvironmentDelete(command.Lister):
    """Delete an environment."""

    def get_parser(self, prog_name):
        parser = super(EnvironmentDelete, self).get_parser(prog_name)
        parser.add_argument(
            'id',
            metavar="<NAME or ID>",
            nargs="+",
            help="Id or name of environment(s) to delete.",
        )
        parser.add_argument(
            '--abandon',
            action='store_true',
            default=False,
            help="If set will abandon environment without deleting any of its"
                 " resources.",
        )

        return parser

    def take_action(self, parsed_args):
        LOG.debug("take_action({0})".format(parsed_args))
        client = self.app.client_manager.application_catalog

        abandon = getattr(parsed_args, 'abandon', False)
        failure_count = 0
        for environment_id in parsed_args.id:
            try:
                environment = murano_utils.find_resource(client.environments,
                                                         environment_id)
                client.environments.delete(environment.id, abandon)
            except exceptions.NotFound:
                failure_count += 1
                print("Failed to delete '{0}'; environment not found".
                      format(environment_id))

        if failure_count == len(parsed_args.id):
            raise exceptions.CommandError("Unable to find and delete any of "
                                          "the specified environments.")

        data = client.environments.list()

        columns = ('id', 'name', 'status', 'created', 'updated')
        column_headers = [c.capitalize() for c in columns]

        return (
            column_headers,
            list(utils.get_item_properties(
                s,
                columns,
            ) for s in data)
        )


class EnvironmentDeploy(command.ShowOne):
    """Start deployment of a murano environment session."""

    def get_parser(self, prog_name):
        parser = super(EnvironmentDeploy, self).get_parser(prog_name)
        parser.add_argument(
            'id',
            metavar="<ENVIRONMENT_ID>",
            help="ID of Environment to deploy.",
        )
        parser.add_argument(
            '--session-id',
            metavar="<SESSION>",
            help="ID of configuration session to deploy.",
        )

        return parser

    def take_action(self, parsed_args):
        LOG.debug("take_action({0})".format(parsed_args))
        client = self.app.client_manager.application_catalog

        client.sessions.deploy(parsed_args.id, parsed_args.session_id)

        environment = utils.find_resource(client.environments,
                                          parsed_args.id)
        data = client.environments.get(environment.id,
                                       parsed_args.session_id).to_dict()

        data['services'] = jsonutils.dumps(data['services'], indent=2)

        if getattr(parsed_args, 'only_apps', False):
            return(['services'], [data['services']])
        else:
            return self.dict2columns(data)


class EnvironmentAppsEdit(command.Command):
    """Edit environment's services list.

    `FILE` is path to a file, that contains jsonpatch, that describes changes
    to be made to environment's object-model.

    [
        { "op": "add", "path": "/-",
           "value": { ... your-app object model here ... }
        },
        { "op": "replace", "path": "/0/?/name",
          "value": "new_name"
        },
    ]

    NOTE: Values '===id1===', '===id2===', etc. in the resulting object-model
    will be substituted with uuids.

    For more info on jsonpatch see RFC 6902
    """

    def get_parser(self, prog_name):
        parser = super(EnvironmentAppsEdit, self).get_parser(prog_name)
        parser.add_argument(
            'id',
            metavar="<ENVIRONMENT_ID>",
            help="ID of Environment to edit.",
        )
        parser.add_argument(
            'filename',
            metavar="<FILE>",
            help="File to read jsonpatch from (defaults to stdin).",
        )
        parser.add_argument(
            '--session-id',
            metavar="<SESSION>",
            help="ID of configuration session to edit.",
        )

        return parser

    def take_action(self, parsed_args):
        LOG.debug("take_action(%s)", parsed_args)
        client = self.app.client_manager.application_catalog
        jp_obj = None
        if not parsed_args.filename:
            jp_obj = json.load(sys.stdin)
        else:
            with open(parsed_args.filename) as fpatch:
                jp_obj = json.load(fpatch)

        jpatch = jsonpatch.JsonPatch(jp_obj)
        environment_id = parsed_args.id
        session_id = parsed_args.session_id
        environment = client.environments.get(environment_id, session_id)

        object_model = jpatch.apply(environment.services)
        murano_utils.traverse_and_replace(object_model)

        client.services.put(
            environment_id,
            path='/',
            data=jpatch.apply(environment.services),
            session_id=session_id)


class EnvironmentModelShow(command.ShowOne):
    """Show environment's object model."""

    def get_parser(self, prog_name):
        parser = super(EnvironmentModelShow, self).get_parser(prog_name)
        parser.add_argument(
            'id',
            metavar="<ENVIRONMENT_ID>",
            help="ID of Environment to show.",
        )
        parser.add_argument(
            "--path",
            metavar="<PATH>",
            default='/',
            help="Path to Environment model section. Defaults to '/'."
        )
        parser.add_argument(
            '--session-id',
            metavar="<SESSION_ID>",
            help="Id of a config session.",
        )

        return parser

    def take_action(self, parsed_args):
        LOG.debug("take_action(%s)", parsed_args)
        client = self.app.client_manager.application_catalog

        session_id = parsed_args.session_id or None
        path = urllib.parse.quote(parsed_args.path)
        env_model = client.environments.get_model(parsed_args.id, path,
                                                  session_id)

        return self.dict2columns(env_model)


class EnvironmentModelEdit(command.ShowOne):
    """Edit environment's object model."""

    def get_parser(self, prog_name):
        parser = super(EnvironmentModelEdit, self).get_parser(prog_name)
        parser.add_argument(
            'id',
            metavar="<ENVIRONMENT_ID>",
            help="ID of Environment to edit.",
        )
        parser.add_argument(
            "filename",
            metavar="<FILE>",
            nargs="?",
            help="File to read JSON-patch from (defaults to stdin)."
        )
        parser.add_argument(
            '--session-id',
            metavar="<SESSION_ID>",
            help="Id of a config session.",
        )

        return parser

    def take_action(self, parsed_args):
        LOG.debug("take_action(%s)", parsed_args)
        client = self.app.client_manager.application_catalog

        jp_obj = None
        if not parsed_args.filename:
            jp_obj = json.load(sys.stdin)
        else:
            with open(parsed_args.filename) as fpatch:
                jp_obj = json.load(fpatch)

        if not isinstance(jp_obj, list):
            raise exceptions.CommandError(
                'JSON-patch must be a list of changes')
        for change in jp_obj:
            if 'op' not in change or 'path' not in change:
                raise exceptions.CommandError(
                    'Every change in JSON-patch must contain "op" and "path" '
                    'keys')
            op = change['op']
            if op not in ['add', 'replace', 'remove']:
                raise exceptions.CommandError('The value of "op" item must be '
                                              '"add", "replace" or "remove", '
                                              'got {0}'.format(op))
            if op != 'remove' and 'value' not in change:
                raise exceptions.CommandError('"add" or "replace" change in '
                                              'JSON-patch must contain "value"'
                                              ' key')
        session_id = parsed_args.session_id
        new_model = client.environments.update_model(parsed_args.id, jp_obj,
                                                     session_id)

        return self.dict2columns(new_model)



# Copyright 2015 Huawei.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools

from muranoclient.common import base
from muranoclient.v1 import packages


class BaseTest(testtools.TestCase):

    def test_two_resources_with_same_id_are_not_equal(self):
        # Two resources with same ID: never equal if their info is not equal
        r1 = base.Resource(None, {'id': 1, 'name': 'hi'})
        r2 = base.Resource(None, {'id': 1, 'name': 'hello'})
        self.assertNotEqual(r1, r2)

    def test_two_resources_with_same_id_and_info_are_equal(self):
        # Two resources with same ID: equal if their info is equal
        r1 = base.Resource(None, {'id': 1, 'name': 'hello'})
        r2 = base.Resource(None, {'id': 1, 'name': 'hello'})
        self.assertEqual(r1, r2)

    def test_two_resources_with_diff_type_are_not_equal(self):
        # Two resources of different types: never equal
        r1 = base.Resource(None, {'id': 1})
        r2 = packages.Package(None, {'id': 1})
        self.assertNotEqual(r1, r2)

    def test_two_resources_with_no_id_are_equal(self):
        # Two resources with no ID: equal if their info is equal
        r1 = base.Resource(None, {'name': 'joe', 'age': 12})
        r2 = base.Resource(None, {'name': 'joe', 'age': 12})
        self.assertEqual(r1, r2)



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from muranoclient.common import base


class RequestStatistics(base.Resource):
    def __repr__(self):
        return "<Request statistics %s>" % self._info

    def data(self, **kwargs):
        return self.manager.data(self, **kwargs)


class RequestStatisticsManager(base.Manager):
    resource_class = RequestStatistics

    def list(self):
        return self._list('/v1/stats')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import json
import tempfile
from unittest import mock


from muranoclient.osc.v1 import environment as osc_env
from muranoclient.tests.unit.osc.v1 import fakes
from muranoclient.v1 import environments as api_env

ENV_INFO = {'id': '1234',
            'name': 'Fake Environment',
            'created': '2015-12-16T17:31:54',
            'updated': '2015-12-16T17:31:54',
            'networking': {},
            'services': ['fake services'],
            'status': 'fake deployed',
            'tenant_id': 'xyz123',
            'version': '1'}

ENV_MODEL = {
    "defaultNetworks": {
        "environment": {
            "name": "env-network",
            "?": {
                "type": "io.murano.resources.NeutronNetwork",
                "id": "5678"
            }
        },
        "flat": None
    },
    "region": "RegionOne",
    "name": "env",
    "?": {
        "updated": "2016-10-03 09:33:41.039789",
        "type": "io.murano.Environment",
        "id": "1234"
    }
}


class TestEnvironment(fakes.TestApplicationCatalog):
    def setUp(self):
        super(TestEnvironment, self).setUp()
        self.environment_mock = self.app.client_manager.application_catalog.\
            environments
        self.session_mock = self.app.client_manager.application_catalog.\
            sessions
        self.services_mock = self.app.client_manager.application_catalog.\
            services
        self.environment_mock.reset_mock()


class TestListEnvironment(TestEnvironment):
    def setUp(self):
        super(TestListEnvironment, self).setUp()
        self.environment_mock.list.return_value = [api_env.Environment(None,
                                                   ENV_INFO)]

        # Command to test
        self.cmd = osc_env.ListEnvironments(self.app, None)

    @mock.patch('osc_lib.utils.get_item_properties')
    def test_environment_list_with_no_options(self, mock_util):
        arglist = []
        verifylist = []

        mock_util.return_value = ('1234', 'Environment of all tenants',
                                  'fake deployed', '2015-12-16T17:31:54',
                                  '2015-12-16T17:31:54'
                                  )

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = ['Id', 'Name', 'Status', 'Created', 'Updated']
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        expected_data = [('1234', 'Environment of all tenants',
                          'fake deployed', '2015-12-16T17:31:54',
                          '2015-12-16T17:31:54')]
        self.assertEqual(expected_data, data)

    @mock.patch('osc_lib.utils.get_item_properties')
    def test_environment_list_with_all_tenants(self, mock_util):
        arglist = ['--all-tenants']
        verifylist = [('all_tenants', True), ('tenant', None)]

        mock_util.return_value = ('1234', 'Environment of all tenants',
                                  'fake deployed', '2015-12-16T17:31:54',
                                  '2015-12-16T17:31:54'
                                  )

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = ['Id', 'Name', 'Status', 'Created', 'Updated']
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        expected_data = [('1234', 'Environment of all tenants',
                          'fake deployed', '2015-12-16T17:31:54',
                          '2015-12-16T17:31:54')]
        self.assertEqual(expected_data, data)
        self.environment_mock.list.assert_called_once_with(True, None)

    @mock.patch('osc_lib.utils.get_item_properties')
    def test_environment_list_with_tenant(self, mock_util):
        arglist = ['--tenant=ABC']
        verifylist = [('all_tenants', False), ('tenant', 'ABC')]

        mock_util.return_value = ('1234', 'Environment of tenant ABC',
                                  'fake deployed', '2015-12-16T17:31:54',
                                  '2015-12-16T17:31:54'
                                  )

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = ['Id', 'Name', 'Status', 'Created', 'Updated']
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        expected_data = [('1234', 'Environment of tenant ABC',
                          'fake deployed', '2015-12-16T17:31:54',
                          '2015-12-16T17:31:54')]
        self.assertEqual(expected_data, data)
        self.environment_mock.list.assert_called_once_with(False, 'ABC')


class TestShowEnvironment(TestEnvironment):
    def setUp(self):
        super(TestShowEnvironment, self).setUp()
        mock_to_dict = self.environment_mock.get.return_value.to_dict
        mock_to_dict.return_value = ENV_INFO

        self.cmd = osc_env.ShowEnvironment(self.app, None)

    @mock.patch('oslo_serialization.jsonutils.dumps')
    def test_environment_show_with_no_options(self, mock_json):
        arglist = ['fake']
        verifylist = [('id', 'fake')]

        mock_json.return_value = ['fake services']

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = ('created', 'id', 'name', 'networking', 'services',
                            'status', 'tenant_id', 'updated', 'version')
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        expected_data = ('2015-12-16T17:31:54', '1234', 'Fake Environment',
                         {}, ['fake services'], 'fake deployed', 'xyz123',
                         '2015-12-16T17:31:54', '1')
        self.assertEqual(expected_data, data)

    @mock.patch('oslo_serialization.jsonutils.dumps')
    def test_environment_show_with_only_app_option(self, mock_json):
        arglist = ['fake', '--only-apps']
        verifylist = [('id', 'fake'), ('only_apps', True)]

        mock_json.return_value = ['fake services']

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = ['services']
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        expected_data = [['fake services']]
        self.assertEqual(expected_data, data)

    @mock.patch('oslo_serialization.jsonutils.dumps')
    def test_environment_show_with_session_id_option(self, mock_json):
        arglist = ['fake', '--session-id', 'abc123']
        verifylist = [('id', 'fake'), ('session_id', 'abc123')]

        mock_json.return_value = ['fake services']

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = ('created', 'id', 'name', 'networking', 'services',
                            'status', 'tenant_id', 'updated', 'version')
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        expected_data = ('2015-12-16T17:31:54', '1234', 'Fake Environment',
                         {}, ['fake services'], 'fake deployed', 'xyz123',
                         '2015-12-16T17:31:54', '1')
        self.assertEqual(expected_data, data)


class TestRenameEnvironment(TestEnvironment):
    def setUp(self):
        super(TestRenameEnvironment, self).setUp()
        self.environment_mock.update.return_value = [api_env.Environment(None,
                                                     ENV_INFO)]

        # Command to test
        self.cmd = osc_env.RenameEnvironment(self.app, None)

    @mock.patch('osc_lib.utils.get_item_properties')
    def test_environment_rename(self, mock_util):
        arglist = ['1234', 'fake-1']
        verifylist = [('id', '1234'), ('name', 'fake-1')]

        mock_util.return_value = ('1234', 'fake-1', 'fake deployed',
                                  '2015-12-16T17:31:54', '2015-12-16T17:31:54'
                                  )

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = ['Id', 'Name', 'Status', 'Created', 'Updated']
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        expected_data = [('1234', 'fake-1', 'fake deployed',
                          '2015-12-16T17:31:54', '2015-12-16T17:31:54')]
        self.assertEqual(expected_data, data)


class TestEnvironmentSessionCreate(TestEnvironment):
    def setUp(self):
        super(TestEnvironmentSessionCreate, self).setUp()

        # Command to test
        self.cmd = osc_env.EnvironmentSessionCreate(self.app, None)

    @mock.patch('muranoclient.common.utils.text_wrap_formatter')
    def test_environment_session_create(self, mock_util):
        arglist = ['1234']
        verifylist = [('id', '1234')]

        mock_util.return_value = '1abc2xyz'

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = ['id']
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        expected_data = ['1abc2xyz']
        self.assertEqual(expected_data, data)


class TestEnvironmentCreate(TestEnvironment):
    def setUp(self):
        super(TestEnvironmentCreate, self).setUp()
        self.environment_mock.create.return_value = [api_env.Environment(None,
                                                     ENV_INFO)]

        # Command to test
        self.cmd = osc_env.EnvironmentCreate(self.app, None)

    @mock.patch('osc_lib.utils.get_item_properties')
    def test_environment_create_with_no_option(self, mock_util):
        arglist = ['fake']
        verifylist = [('name', 'fake')]

        mock_util.return_value = ('1234', 'fake', 'ready',
                                  '2015-12-16T17:31:54', '2015-12-16T17:31:54')

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = ['Id', 'Name', 'Status', 'Created', 'Updated']
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        expected_data = [('1234', 'fake', 'ready',
                          '2015-12-16T17:31:54', '2015-12-16T17:31:54')]
        self.assertEqual(expected_data, data)

    @mock.patch('osc_lib.utils.get_item_properties')
    def test_environment_create_with_region_option(self, mock_util):
        arglist = ['fake', '--region', 'region_one']
        verifylist = [('name', 'fake'), ('region', 'region_one')]

        mock_util.return_value = ('1234', 'fake', 'ready',
                                  '2015-12-16T17:31:54', '2015-12-16T17:31:54')

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        # Check that correct arguments are passed
        self.environment_mock.create.assert_has_calls([mock.call(
            {'name': 'fake', 'region': 'region_one'})])

        # Check that columns are correct
        expected_columns = ['Id', 'Name', 'Status', 'Created', 'Updated']
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        expected_data = [('1234', 'fake', 'ready',
                          '2015-12-16T17:31:54', '2015-12-16T17:31:54')]
        self.assertEqual(expected_data, data)

    @mock.patch('osc_lib.utils.get_item_properties')
    def test_environment_create_with_net_option(self, mock_util):
        arglist = ['fake', '--join-net-id', 'x1y2z3']
        verifylist = [('name', 'fake'), ('join_net_id', 'x1y2z3')]

        mock_util.return_value = ('1234', 'fake', 'ready',
                                  '2015-12-16T17:31:54', '2015-12-16T17:31:54')

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        expected_call = {
            'defaultNetworks': {
                'environment': {
                    'internalNetworkName': 'x1y2z3',
                    '?': {
                        'type': 'io.murano.resources.ExistingNeutronNetwork',
                        'id': mock.ANY
                    }
                },
                'flat': None
            },
            'name': 'fake',
            'region': None
        }

        # Check that correct arguments are passed
        self.environment_mock.create.assert_called_with(expected_call)

        # Check that columns are correct
        expected_columns = ['Id', 'Name', 'Status', 'Created', 'Updated']
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        expected_data = [('1234', 'fake', 'ready',
                          '2015-12-16T17:31:54', '2015-12-16T17:31:54')]
        self.assertEqual(expected_data, data)

    @mock.patch('osc_lib.utils.get_item_properties')
    def test_environment_create_with_subnet_option(self, mock_util):
        arglist = ['fake', '--join-subnet-id', 'x1y2z3']
        verifylist = [('name', 'fake'), ('join_subnet_id', 'x1y2z3')]

        mock_util.return_value = ('1234', 'fake', 'ready',
                                  '2015-12-16T17:31:54', '2015-12-16T17:31:54')

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        expected_call = {
            'defaultNetworks': {
                'environment': {
                    'internalSubnetworkName': 'x1y2z3',
                    '?': {
                        'type': 'io.murano.resources.ExistingNeutronNetwork',
                        'id': mock.ANY
                    }
                },
                'flat': None
            },
            'name': 'fake',
            'region': None
        }

        # Check that correct arguments are passed
        self.environment_mock.create.assert_called_with(expected_call)

        # Check that columns are correct
        expected_columns = ['Id', 'Name', 'Status', 'Created', 'Updated']
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        expected_data = [('1234', 'fake', 'ready',
                          '2015-12-16T17:31:54', '2015-12-16T17:31:54')]
        self.assertEqual(expected_data, data)


class TestEnvironmentDelete(TestEnvironment):
    def setUp(self):
        super(TestEnvironmentDelete, self).setUp()
        self.environment_mock.delete.return_value = None
        self.environment_mock.list.return_value = [api_env.Environment(None,
                                                   ENV_INFO)]

        # Command to test
        self.cmd = osc_env.EnvironmentDelete(self.app, None)

    @mock.patch('osc_lib.utils.get_item_properties')
    def test_environment_delete(self, mock_util):
        arglist = ['fake1', 'fake2']
        verifylist = [('id', ['fake1', 'fake2'])]

        mock_util.return_value = ('1234', 'Environment of all tenants',
                                  'fake deployed', '2015-12-16T17:31:54',
                                  '2015-12-16T17:31:54'
                                  )

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = ['Id', 'Name', 'Status', 'Created', 'Updated']
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        expected_data = [('1234', 'Environment of all tenants',
                          'fake deployed', '2015-12-16T17:31:54',
                          '2015-12-16T17:31:54')]
        self.assertEqual(expected_data, data)


class TestEnvironmentDeploy(TestEnvironment):
    def setUp(self):
        super(TestEnvironmentDeploy, self).setUp()
        mock_to_dict = self.environment_mock.get.return_value.to_dict
        mock_to_dict.return_value = ENV_INFO

        # Command to test
        self.cmd = osc_env.EnvironmentDeploy(self.app, None)

    @mock.patch('oslo_serialization.jsonutils.dumps')
    def test_environment_deploy(self, mock_json):
        arglist = ['fake', '--session-id', 'abc123']
        verifylist = [('id', 'fake'), ('session_id', 'abc123')]

        mock_json.return_value = ['fake services']

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = ('created', 'id', 'name', 'networking', 'services',
                            'status', 'tenant_id', 'updated', 'version')
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        expected_data = ('2015-12-16T17:31:54', '1234', 'Fake Environment',
                         {}, ['fake services'], 'fake deployed', 'xyz123',
                         '2015-12-16T17:31:54', '1')
        self.assertEqual(expected_data, data)


class TestEnvironmentAppsEdit(TestEnvironment):
    def setUp(self):
        super(TestEnvironmentAppsEdit, self).setUp()

        # Command to test
        self.cmd = osc_env.EnvironmentAppsEdit(self.app, None)

    def test_environment_apps_edit(self):
        fake = collections.namedtuple('fakeEnv', 'services')
        self.environment_mock.get.side_effect = [
            fake(services=[
                {'?': {'name': "foo"}}
            ]),
        ]

        temp_file = tempfile.NamedTemporaryFile(prefix="murano-test", mode='w')
        json.dump([
            {'op': 'replace', 'path': '/0/?/name',
                'value': "dummy"
             }
        ], temp_file)
        temp_file.file.flush()

        arglist = ['fake', '--session-id', 'abc123', temp_file.name]

        parsed_args = self.check_parser(self.cmd, arglist, [])

        self.cmd.take_action(parsed_args)

        self.services_mock.put.assert_called_once_with(
            'fake',
            session_id='abc123',
            path='/',
            data=[{'?': {'name': 'dummy'}}]
        )


class TestEnvironmentModelShow(TestEnvironment):
    def setUp(self):
        super(TestEnvironmentModelShow, self).setUp()
        self.env_mock = \
            self.app.client_manager.application_catalog.environments
        self.env_mock.get_model.return_value = ENV_MODEL

        # Command to test
        self.cmd = osc_env.EnvironmentModelShow(self.app, None)

    def test_environment_model_show_basic(self):
        arglist = ['env-id']
        verifylist = [('id', 'env-id')]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = ('?', 'defaultNetworks', 'name', 'region')
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        self.assertCountEqual(ENV_MODEL.values(), data)

    def test_environment_model_show_full(self):
        arglist = ['env-id', '--path', '/path', '--session-id', 'sess-id']
        verifylist = [('id', 'env-id'), ('path', '/path'),
                      ('session_id', 'sess-id')]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = ('?', 'defaultNetworks', 'name', 'region')
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        self.assertCountEqual(ENV_MODEL.values(), data)


class TestEnvironmentModelEdit(TestEnvironment):
    def setUp(self):
        super(TestEnvironmentModelEdit, self).setUp()
        self.env_mock = \
            self.app.client_manager.application_catalog.environments
        self.env_mock.update_model.return_value = ENV_MODEL

        # Command to test
        self.cmd = osc_env.EnvironmentModelEdit(self.app, None)

    def test_environment_model_edit(self):
        temp_file = tempfile.NamedTemporaryFile(prefix="murano-test", mode='w')
        patch = [{'op': 'replace', 'path': '/name', 'value': 'dummy'}]
        json.dump(patch, temp_file)
        temp_file.file.flush()

        arglist = ['env-id', temp_file.name, '--session-id', 'sess-id']
        verifylist = [('id', 'env-id'), ('filename', temp_file.name),
                      ('session_id', 'sess-id')]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = ('?', 'defaultNetworks', 'name', 'region')
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        self.assertCountEqual(ENV_MODEL.values(), data)



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import json
import os.path
import tempfile
from unittest import mock
import zipfile

import requests
import requests_mock
import testtools
import yaml

from muranoclient.common import utils


class FileTest(testtools.TestCase):

    def test_file_object_from_file(self):
        f_obj = tempfile.NamedTemporaryFile(delete=True)
        new_f_obj = utils.File(f_obj).open()
        self.assertTrue(hasattr(new_f_obj, 'read'))

        new_f_obj = utils.File(f_obj.name).open()
        self.assertTrue(hasattr(new_f_obj, 'read'))

    def test_file_object_file_fails(self):
        f_obj = utils.File('')
        self.assertRaises(ValueError, f_obj.open)

    def test_file_object_url_fails(self):
        resp = requests.Response()
        resp.status_code = 400
        resp.raw = io.BytesIO(b"123")

        with mock.patch(
                'requests.get',
                mock.Mock(side_effect=lambda k, *args, **kwargs: resp)):
            f = utils.File("http://127.0.0.1")
            self.assertRaises(ValueError, f.open)

    def test_file_object_url(self):
        resp = requests.Response()
        resp.raw = io.BytesIO(b"123")
        resp.status_code = 200
        with mock.patch(
                'requests.get',
                mock.Mock(side_effect=lambda k, *args, **kwargs: resp)):
            new_f_obj = utils.File('http://127.0.0.1/').open()
            self.assertTrue(hasattr(new_f_obj, 'read'))


def make_pkg(manifest_override, image_dicts=None):
    manifest = {
        'Author': '',
        'Classes': {'foo': 'foo.yaml'},
        'Description': '',
        'Format': 1.0,
        'FullName': 'org.foo',
        'Name': 'Apache HTTP Server',
        'Type': 'Application'}
    manifest.update(manifest_override)
    file_obj = io.BytesIO()
    zfile = zipfile.ZipFile(file_obj, "a")
    zfile.writestr('manifest.yaml', yaml.dump(manifest))
    zfile.writestr('Classes/foo.yaml', yaml.dump({}))
    if image_dicts:
        images_list = []
        default_image_spec = {
            'ContainerFormat': 'bare',
            'DiskFormat': 'qcow2',
            'Name': '',
        }
        for image_dict in image_dicts:
            image_spec = default_image_spec.copy()
            image_spec.update(image_dict)
            images_list.append(image_spec)
        images = {'Images': images_list, }
        zfile.writestr('images.lst', yaml.dump(images))
    zfile.close()
    file_obj.seek(0)
    return file_obj


class PackageTest(testtools.TestCase):
    base_url = "http://127.0.0.1"

    @requests_mock.mock()
    def test_from_location_local_file(self, m):
        temp = tempfile.NamedTemporaryFile()
        pkg = make_pkg({'FullName': 'single_app'})

        temp.write(pkg.read())
        temp.flush()
        path, name = os.path.split(temp.name)

        # ensure we do not go to base url
        m.get(self.base_url + '/apps/{0}.zip'.format(name), status_code=404)

        self.assertEqual('single_app', utils.Package.from_location(
            name=name, base_url=self.base_url,
            path=path,
        ).manifest['FullName'])

    def test_package_from_directory(self):
        path = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                            "fixture_data/empty-app")
        pkg = utils.Package(utils.File(path))
        self.assertEqual('empty', pkg.manifest['FullName'])
        pkg = utils.Package.from_location('', path=path)
        self.assertEqual('empty', pkg.manifest['FullName'])

    @requests_mock.mock()
    def test_from_location_url(self, m):
        """Test that url overrides name specification."""

        pkg = make_pkg({'FullName': 'single_app'})
        m.get('http://127.0.0.2/apps/single_app.zip', body=pkg)
        m.get(self.base_url + '/apps/single_app.zip', status_code=404)

        self.assertEqual('single_app', utils.Package.from_location(
            name='single_app', base_url=self.base_url,
            url="http://127.0.0.2/apps/single_app.zip",
        ).manifest['FullName'])

    @requests_mock.mock()
    def test_from_location(self, m):
        """Test from location url requesting mechanism."""
        pkg = make_pkg({'FullName': 'single_app'})
        pkg_ver = make_pkg({'FullName': 'single_app'})
        m.get(self.base_url + '/apps/single_app.zip', body=pkg)
        m.get(self.base_url + '/apps/single_app.1.0.zip', body=pkg_ver)
        m.get(self.base_url + '/apps/single_app.2.0.zip', status_code=404)

        self.assertEqual('single_app', utils.Package.from_location(
            name='single_app', base_url=self.base_url).manifest['FullName'])

        self.assertEqual('single_app', utils.Package.from_location(
            name='single_app',
            version='1.0',
            base_url=self.base_url).manifest['FullName'])
        self.assertRaises(
            ValueError,
            utils.Package.from_location,
            name='single_app',
            version='2.0',
            base_url=self.base_url)

    def test_no_requirements(self):
        pkg = make_pkg({'FullName': 'single_app'})
        app = utils.Package.fromFile(pkg)
        reqs = app.requirements(base_url=self.base_url)
        self.assertEqual({'single_app': app}, reqs)

    @requests_mock.mock()
    def test_requirements(self, m):
        """Test that dependencies are parsed correctly."""

        pkg3 = make_pkg({'FullName': 'dep_of_dep'})
        pkg2 = make_pkg({'FullName': 'dep_app', 'Require': {
            'dep_of_dep': "1.0"}, })
        pkg1 = make_pkg({'FullName': 'main_app', 'Require': {
            'dep_app': None}, })

        m.get(self.base_url + '/apps/main_app.zip', body=pkg1)
        m.get(self.base_url + '/apps/dep_app.zip', body=pkg2)
        m.get(self.base_url + '/apps/dep_of_dep.1.0.zip', body=pkg3)
        app = utils.Package.fromFile(pkg1)
        reqs = app.requirements(base_url=self.base_url)

        self.assertEqual(
            {'main_app': app, 'dep_app': mock.ANY, 'dep_of_dep': mock.ANY},
            reqs)

    @mock.patch('muranoclient.common.utils.Package.from_file')
    def test_requirements_order(self, from_file):
        """Test that dependencies are parsed in correct order."""

        pkg5 = make_pkg({'FullName': 'd4', })
        pkg4 = make_pkg({'FullName': 'd3', 'Require': {'d4': None}, })
        pkg3 = make_pkg({'FullName': 'd2', 'Require': {'d3': None}, })
        pkg2 = make_pkg({'FullName': 'd1', 'Require': {'d3': None}, })
        pkg1 = make_pkg({'FullName': 'M', 'Require': {'d1': None,
                                                      'd2': None,
                                                      'd4': None}, })

        def side_effect(name):
            if 'M' in name:
                return utils.Package(utils.File(pkg1))
            if 'd1' in name:
                return utils.Package(utils.File(pkg2))
            if 'd2' in name:
                return utils.Package(utils.File(pkg3))
            if 'd3' in name:
                return utils.Package(utils.File(pkg4))
            if 'd4' in name:
                return utils.Package(utils.File(pkg5))

        from_file.side_effect = side_effect
        app = from_file('M')
        reqs = app.requirements(base_url=self.base_url)

        def key_position(key):
            keys = list(iter(reqs.keys()))
            return keys.index(key)

        self.assertTrue(
            key_position('d4') < key_position('d3') and
            key_position('d4') < key_position('M') and
            key_position('d3') < key_position('d1') and
            key_position('d3') < key_position('d2') < key_position('M')
        )

    @mock.patch('muranoclient.common.utils.Package.from_file')
    def test_requirements_order2(self, from_file):
        """Test that dependencies are parsed in correct order."""

        pkg5 = make_pkg({'FullName': 'd4', 'Require': {'d6': None}, })
        pkg4 = make_pkg({'FullName': 'd3', 'Require': {'d4': None}, })
        pkg3 = make_pkg({'FullName': 'd1', 'Require': {'d3': None,
                                                       'd7': None}, })
        pkg2 = make_pkg({'FullName': 'd2', 'Require': {'d3': None}, })
        pkg6 = make_pkg({'FullName': 'd6', })
        pkg7 = make_pkg({'FullName': 'd7', 'Require': {'d8': None}, })
        pkg8 = make_pkg({'FullName': 'd8', })

        pkg1 = make_pkg({'FullName': 'M', 'Require': {'d1': None,
                                                      'd2': None,
                                                      'd4': None, }, })

        def side_effect(name):
            if 'M' in name:
                return utils.Package(utils.File(pkg1))
            if 'd1' in name:
                return utils.Package(utils.File(pkg2))
            if 'd2' in name:
                return utils.Package(utils.File(pkg3))
            if 'd3' in name:
                return utils.Package(utils.File(pkg4))
            if 'd4' in name:
                return utils.Package(utils.File(pkg5))
            if 'd6' in name:
                return utils.Package(utils.File(pkg6))
            if 'd7' in name:
                return utils.Package(utils.File(pkg7))
            if 'd8' in name:
                return utils.Package(utils.File(pkg8))

        from_file.side_effect = side_effect
        app = from_file('M')
        reqs = app.requirements(base_url=self.base_url)

        def key_position(key):
            keys = list(iter(reqs.keys()))
            return keys.index(key)

        self.assertTrue(
            key_position('d6') < key_position('d4') <
            key_position('d3') < key_position('d1') and
            key_position('d3') < key_position('d2') and
            key_position('d1') < key_position('M') and
            key_position('d2') < key_position('M') and
            key_position('d8') < key_position('d7') < key_position('d1')
        )

    @mock.patch('muranoclient.common.utils.Package.from_file')
    def test_cyclic_requirements(self, from_file):
        """Test that a cyclic dependency would be handled correctly."""
        pkg3 = make_pkg({'FullName': 'dep_of_dep', 'Require': {
            'main_app': None, 'dep_app': None}, })
        pkg2 = make_pkg({'FullName': 'dep_app', 'Require': {
            'dep_of_dep': None, 'main_app': None}, })
        pkg1 = make_pkg({'FullName': 'main_app', 'Require': {
            'dep_app': None, 'dep_of_dep': None}, })

        def side_effect(name):
            if 'main_app' in name:
                return utils.Package(utils.File(pkg1))
            if 'dep_app' in name:
                return utils.Package(utils.File(pkg2))
            if 'dep_of_dep' in name:
                return utils.Package(utils.File(pkg3))

        from_file.side_effect = side_effect
        app = from_file('main_app')
        reqs = app.requirements(base_url=self.base_url)

        self.assertEqual(
            {'main_app': app, 'dep_app': mock.ANY, 'dep_of_dep': mock.ANY},
            reqs)

    @mock.patch('muranoclient.common.utils.Package.from_file')
    def test_order_with_cyclic_requirements2(self, from_file):
        """Test that dependencies are parsed in correct order."""

        pkg6 = make_pkg({'FullName': 'd5', 'Require': {'d6': None}, })
        pkg7 = make_pkg({'FullName': 'd6', })
        pkg5 = make_pkg({'FullName': 'd4', 'Require': {'d3': None,
                                                       'd5': None}})
        pkg4 = make_pkg({'FullName': 'd3', 'Require': {'d4': None}, })
        pkg3 = make_pkg({'FullName': 'd2', 'Require': {'d1': None,
                                                       'd5': None,
                                                       'd6': None}, })
        pkg2 = make_pkg({'FullName': 'd1', 'Require': {'d2': None}, })
        pkg1 = make_pkg({'FullName': 'M', 'Require': {'d1': None,
                                                      'd3': None}, })

        def side_effect(name):
            if 'M' in name:
                return utils.Package(utils.File(pkg1))
            if 'd1' in name:
                return utils.Package(utils.File(pkg2))
            if 'd2' in name:
                return utils.Package(utils.File(pkg3))
            if 'd3' in name:
                return utils.Package(utils.File(pkg4))
            if 'd4' in name:
                return utils.Package(utils.File(pkg5))
            if 'd5' in name:
                return utils.Package(utils.File(pkg6))
            if 'd6' in name:
                return utils.Package(utils.File(pkg7))

        from_file.side_effect = side_effect
        app = from_file('M')
        reqs = app.requirements(base_url=self.base_url)

        def key_position(key):
            keys = list(iter(reqs.keys()))
            return keys.index(key)

        self.assertTrue(
            key_position('d5') < key_position('d4') and
            key_position('d5') < key_position('d2') and
            key_position('d5') < key_position('d3') < key_position('M') and
            key_position('d5') < key_position('d1') < key_position('M')
        )

    @mock.patch('muranoclient.common.utils.Package.from_file')
    def test_order_with_cyclic_requirements3(self, from_file):
        """Test that dependencies are parsed in correct order."""

        pkg5 = make_pkg({'FullName': 'd4', })
        pkg4 = make_pkg({'FullName': 'd3', 'Require': {'M': None}, })
        pkg3 = make_pkg({'FullName': 'd2', 'Require': {'d3': None,
                                                       'd4': None}, })
        pkg2 = make_pkg({'FullName': 'd1', 'Require': {'d2': None}, })
        pkg1 = make_pkg({'FullName': 'M', 'Require': {'d1': None}, })

        def side_effect(name):
            if 'M' in name:
                return utils.Package(utils.File(pkg1))
            if 'd1' in name:
                return utils.Package(utils.File(pkg2))
            if 'd2' in name:
                return utils.Package(utils.File(pkg3))
            if 'd3' in name:
                return utils.Package(utils.File(pkg4))
            if 'd4' in name:
                return utils.Package(utils.File(pkg5))

        from_file.side_effect = side_effect
        app = from_file('M')
        reqs = app.requirements(base_url=self.base_url)

        def key_position(key):
            keys = list(iter(reqs.keys()))
            return keys.index(key)

        self.assertTrue(
            key_position('d4') < key_position('M') and
            key_position('d4') < key_position('d1') and
            key_position('d4') < key_position('d2') and
            key_position('d4') < key_position('d3')
        )

    def test_images(self):
        pkg = make_pkg({})
        app = utils.Package.fromFile(pkg)
        self.assertEqual([], app.images())

        pkg = make_pkg(
            {}, [{'Name': 'test.qcow2'}, {'Name': 'test2.qcow2'}])
        app = utils.Package.fromFile(pkg)
        self.assertEqual(
            set(['test.qcow2', 'test2.qcow2']),
            set([img['Name'] for img in app.images()]))

    def test_file_object_repo_fails(self):
        resp = requests.Response()
        resp.raw = io.BytesIO(b"123")
        resp.status_code = 400
        with mock.patch(
                'requests.get',
                mock.Mock(side_effect=lambda k, *args, **kwargs: resp)):
            self.assertRaises(
                ValueError, utils.Package.from_location,
                name='foo.bar.baz', base_url='http://127.0.0.1')

    def test_no_repo_url_fails(self):
        self.assertRaises(ValueError, utils.Package.from_location,
                          name='foo.bar.baz', base_url='')

    @mock.patch.object(utils.Package, 'validate')
    def test_file_object_repo(self, m_validate):
        resp = requests.Response()
        resp.raw = io.BytesIO(b"123")
        resp.status_code = 200
        m_validate.return_value = None
        with mock.patch(
                'requests.get',
                mock.Mock(side_effect=lambda k, *args, **kwargs: resp)):
            new_f_obj = utils.Package.from_location(
                name='foo.bar.baz', base_url='http://127.0.0.1').file()
            self.assertTrue(hasattr(new_f_obj, 'read'))


class BundleTest(testtools.TestCase):
    base_url = "http://127.0.0.1"

    @requests_mock.mock()
    def test_packages(self, m):
        s = io.StringIO()
        bundle_contents = {'Packages': [
            {'Name': 'first_app'},
            {'Name': 'second_app', 'Version': '1.0'}
        ]}
        json.dump(bundle_contents, s)
        s.seek(0)
        bundle = utils.Bundle.from_file(s)
        self.assertEqual(
            set(['first_app', 'second_app']),
            set([p['Name'] for p in bundle.package_specs()])
        )

        # setup packages
        pkg1 = make_pkg({'FullName': 'first_app'})
        pkg2 = make_pkg({'FullName': 'second_app'})

        m.get(self.base_url + '/apps/first_app.zip', body=pkg1)
        m.get(self.base_url + '/apps/second_app.1.0.zip', body=pkg2)
        self.assertEqual(
            set(['first_app', 'second_app']),
            set([p.manifest['FullName']
                 for p in
                 bundle.packages(base_url=self.base_url)])
        )


class TraverseTest(testtools.TestCase):

    def test_traverse_and_replace(self):
        obj = [
            {'id': '===id1==='},
            {'id': '===id2===', 'x': [{'bar': '===id1==='}]},
            ['===id1===', '===id2==='],
            '===id3===',
            '===nonid0===',
            '===id3===',
        ]
        utils.traverse_and_replace(obj)
        self.assertNotEqual('===id1===', obj[0]['id'])
        self.assertNotEqual('===id2===', obj[1]['id'])
        self.assertNotEqual('===id1===', obj[1]['x'][0]['bar'])
        self.assertNotEqual('===id1===', obj[2][0])
        self.assertNotEqual('===id2===', obj[2][1])
        self.assertNotEqual('===id3===', obj[3])
        self.assertEqual('===nonid0===', obj[4])
        self.assertNotEqual('===id3===', obj[5])

        self.assertEqual(obj[0]['id'], obj[1]['x'][0]['bar'])
        self.assertEqual(obj[0]['id'], obj[2][0])

        self.assertEqual(obj[1]['id'], obj[2][1])

        self.assertEqual(obj[3], obj[5])



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from muranoclient.common import base


class Deployment(base.Resource):
    def __repr__(self):
        return '<Deployment %s>' % self._info

    def data(self, **kwargs):
        return self.manager.data(self, **kwargs)


class Status(base.Resource):
    def __repr__(self):
        return '<Status %s>' % self._info

    def data(self, **kwargs):
        return self.manager.data(self, **kwargs)


class DeploymentManager(base.Manager):
    resource_class = Deployment

    def list(self, environment_id, all_environments=False):
        if all_environments:
            return self._list('/v1/deployments', 'deployments')
        else:
            return self._list('/v1/environments/{id}/deployments'.
                              format(id=environment_id), 'deployments')

    def reports(self, environment_id, deployment_id, *service_ids):
        path = '/v1/environments/{id}/deployments/{deployment_id}'
        path = path.format(id=environment_id, deployment_id=deployment_id)
        if service_ids:
            for service_id in service_ids:
                path += '?service_id={0}'.format(service_id)

        resp, body = self.api.json_request(path, 'GET')

        data = body.get('reports', [])
        return [Status(self, res, loaded=True) for res in data if res]



# -*- coding: utf-8 -*-
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import os

# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc',
              'openstackdocstheme',]

# autodoc generation is a bit aggressive and a nuisance when doing heavy
# text edit cycles.
# execute "export SPHINX_DEBUG=1" in your terminal to disable

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
copyright = u'OpenStack Foundation'
exclude_trees = ['api']

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'native'

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
html_theme = 'openstackdocs'

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = ['_theme']
#html_theme_path = [openstackdocstheme.get_html_theme_path()]

# openstackdocstheme options
openstackdocs_repo_name = 'openstack/python-muranoclient'
openstackdocs_bug_project = 'python-muranoclient'
openstackdocs_bug_tag = ''

# Output file base name for HTML help builder.
htmlhelp_basename = 'python-muranoclientdoc'


# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_documents = [
    (
        'index',
        'python-muranoclient.tex',
        u'python-muranoclient Documentation',
        u'OpenStack Foundation',
        'manual'
    ),
]

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'http://docs.python.org/': None}



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Not a true manager yet; should be changed to be one if CRUD
# functionality becomes available for actions.
class ActionManager(object):
    def __init__(self, api):
        self.api = api

    def call(self, environment_id, action_id, arguments=None):
        if arguments is None:
            arguments = {}
        url = '/v1/environments/{environment_id}/actions/{action_id}'.format(
            environment_id=environment_id, action_id=action_id)
        resp, body = self.api.json_request(url, 'POST', body=arguments)
        return body['task_id']

    def get_result(self, environment_id, task_id):
        url = '/v1/environments/{environment_id}/actions/{task_id}'.format(
            environment_id=environment_id, task_id=task_id)
        resp, body = self.api.json_request(url, 'GET')
        return body or None



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Command-line interface to the Murano Project.
"""

import argparse
import sys

import glanceclient
from keystoneclient.auth.identity.generic.cli import DefaultCLI
from keystoneclient.auth.identity import v3 as identity
from keystoneclient import discover
from keystoneclient import exceptions as ks_exc
from keystoneclient import session as ksession
from oslo_log import handlers
from oslo_log import log as logging
from oslo_utils import encodeutils
from oslo_utils import importutils
import urllib.parse as urlparse

import muranoclient
from muranoclient.apiclient import exceptions as exc
from muranoclient import client as murano_client
from muranoclient.common import utils
from muranoclient.glance import client as art_client


logger = logging.getLogger(__name__)

DEFAULT_REPO_URL = "http://apps.openstack.org/api/v1/murano_repo/liberty/"


# quick local fix for keystoneclient bug which blocks built-in reauth
# functionality in case of expired token.
# bug: https://bugs.launchpad.net/python-keystoneclient/+bug/1551392
# fix: https://review.opendev.org/#/c/286236/
class AuthCLI(DefaultCLI):
    def invalidate(self):
        retval = super(AuthCLI, self).invalidate()
        if self._token:
            self._token = None
            retval = True
        return retval


class MuranoShell(object):

    def _append_global_identity_args(self, parser):
        # Register the CLI arguments that have moved to the session object.
        ksession.Session.register_cli_options(parser)

        identity.Password.register_argparse_arguments(parser)

    def get_base_parser(self, argv):

        parser = argparse.ArgumentParser(
            prog='murano',
            description=__doc__.strip(),
            epilog='See "murano help COMMAND" '
                   'for help on a specific command.',
            add_help=False,
            formatter_class=HelpFormatter,
        )

        # Global arguments
        parser.add_argument('-h', '--help',
                            action='store_true',
                            help=argparse.SUPPRESS, )

        parser.add_argument('--version',
                            action='version',
                            version=muranoclient.__version__,
                            help="Show program's version number and exit.")

        parser.add_argument('-d', '--debug',
                            default=bool(utils.env('MURANOCLIENT_DEBUG')),
                            action='store_true',
                            help='Defaults to env[MURANOCLIENT_DEBUG].')

        parser.add_argument('-v', '--verbose',
                            default=False, action="store_true",
                            help="Print more verbose output.")

        parser.add_argument('--api-timeout',
                            help='Number of seconds to wait for an '
                                 'API response, '
                                 'defaults to system socket timeout.')

        parser.add_argument('--os-tenant-id',
                            default=utils.env('OS_TENANT_ID'),
                            help='Defaults to env[OS_TENANT_ID].')

        parser.add_argument('--os-tenant-name',
                            default=utils.env('OS_TENANT_NAME'),
                            help='Defaults to env[OS_TENANT_NAME].')

        parser.add_argument('--os-region-name',
                            default=utils.env('OS_REGION_NAME'),
                            help='Defaults to env[OS_REGION_NAME].')

        parser.add_argument('--os-auth-token',
                            default=utils.env('OS_AUTH_TOKEN'),
                            help='Defaults to env[OS_AUTH_TOKEN].')

        parser.add_argument('--os-no-client-auth',
                            default=utils.env('OS_NO_CLIENT_AUTH'),
                            action='store_true',
                            help="Do not contact keystone for a token. "
                                 "Defaults to env[OS_NO_CLIENT_AUTH].")
        parser.add_argument('--murano-url',
                            default=utils.env('MURANO_URL'),
                            help='Defaults to env[MURANO_URL].')

        parser.add_argument('--glance-url',
                            default=utils.env('GLANCE_URL'),
                            help='Defaults to env[GLANCE_URL].')

        parser.add_argument('--glare-url',
                            default=utils.env('GLARE_URL'),
                            help='Defaults to env[GLARE_URL].')

        parser.add_argument('--murano-api-version',
                            default=utils.env(
                                'MURANO_API_VERSION', default='1'),
                            help='Defaults to env[MURANO_API_VERSION] '
                                 'or 1.')

        parser.add_argument('--os-service-type',
                            default=utils.env('OS_SERVICE_TYPE'),
                            help='Defaults to env[OS_SERVICE_TYPE].')

        parser.add_argument('--os-endpoint-type',
                            default=utils.env('OS_ENDPOINT_TYPE'),
                            help='Defaults to env[OS_ENDPOINT_TYPE].')

        parser.add_argument('--include-password',
                            default=bool(utils.env('MURANO_INCLUDE_PASSWORD')),
                            action='store_true',
                            help='Send os-username and os-password to murano.')

        parser.add_argument('--murano-repo-url',
                            default=utils.env(
                                'MURANO_REPO_URL',
                                default=DEFAULT_REPO_URL),
                            help=('Defaults to env[MURANO_REPO_URL] '
                                  'or {0}'.format(DEFAULT_REPO_URL)))

        parser.add_argument('--murano-packages-service',
                            choices=['murano', 'glance', 'glare'],
                            default=utils.env('MURANO_PACKAGES_SERVICE',
                                              default='murano'),
                            help='Specifies if murano-api ("murano") or '
                                 'Glance Artifact Repository ("glare") '
                                 'should be used to store murano packages. '
                                 'Defaults to env[MURANO_PACKAGES_SERVICE] or '
                                 'to "murano"')

        # The following 3 arguments are deprecated and are all added
        # by keystone session register_cli_opts later.  Only add these
        # arguments if they are present on the command line.

        if '--cert-file' in argv:
            parser.add_argument('--cert-file',
                                dest='os_cert',
                                help='DEPRECATED! Use --os-cert.')

        if '--key-file' in argv:
            parser.add_argument('--key-file',
                                dest='os_key',
                                help='DEPRECATED! Use --os-key.')

        if '--ca-file' in argv:
            parser.add_argument('--ca-file',
                                dest='os_cacert',
                                help='DEPRECATED! Use --os-cacert.')

        self._append_global_identity_args(parser)

        return parser

    def get_subcommand_parser(self, version, argv):
        parser = self.get_base_parser(argv)

        self.subcommands = {}
        subparsers = parser.add_subparsers(metavar='<subcommand>')
        submodule = importutils.import_versioned_module('muranoclient',
                                                        version, 'shell')
        self._find_actions(subparsers, submodule)
        self._find_actions(subparsers, self)

        return parser

    def _find_actions(self, subparsers, actions_module):
        for attr in (a for a in dir(actions_module) if a.startswith('do_')):
            # I prefer to be hypen-separated instead of underscores.
            command = attr[3:].replace('_', '-')
            callback = getattr(actions_module, attr)
            desc = callback.__doc__ or ''
            help = desc.strip().split('\n')[0]
            arguments = getattr(callback, 'arguments', [])

            subparser = subparsers.add_parser(command, help=help,
                                              description=desc,
                                              add_help=False,
                                              formatter_class=HelpFormatter)
            subparser.add_argument('-h', '--help', action='help',
                                   help=argparse.SUPPRESS)
            self.subcommands[command] = subparser
            for (args, kwargs) in arguments:
                subparser.add_argument(*args, **kwargs)
            subparser.set_defaults(func=callback)

    def _discover_auth_versions(self, session, auth_url):
        # discover the API versions the server is supporting base on the
        # given URL
        v2_auth_url = None
        v3_auth_url = None
        try:
            ks_discover = discover.Discover(session=session, auth_url=auth_url)
            v2_auth_url = ks_discover.url_for('2.0')
            v3_auth_url = ks_discover.url_for('3.0')
        except ks_exc.ClientException as e:
            # Identity service may not support discover API version.
            # Lets trying to figure out the API version from the original URL.
            url_parts = urlparse.urlparse(auth_url)
            (scheme, netloc, path, params, query, fragment) = url_parts
            path = path.lower()
            if path.startswith('/v3'):
                v3_auth_url = auth_url
            elif path.startswith('/v2'):
                v2_auth_url = auth_url
            else:
                # not enough information to determine the auth version
                msg = ('Unable to determine the Keystone version '
                       'to authenticate with using the given '
                       'auth_url. Identity service may not support API '
                       'version discovery. Please provide a versioned '
                       'auth_url instead. error=%s') % (e)
                raise exc.CommandError(msg)

        return (v2_auth_url, v3_auth_url)

    def _setup_logging(self, debug):
        # Output the logs to command-line interface
        color_handler = handlers.ColorHandler(sys.stdout)
        logger_root = logging.getLogger(None).logger
        logger_root.level = logging.DEBUG if debug else logging.WARNING
        logger_root.addHandler(color_handler)

        # Set the logger level of special library
        logging.getLogger('iso8601') \
            .logger.setLevel(logging.WARNING)
        logging.getLogger('urllib3.connectionpool') \
            .logger.setLevel(logging.WARNING)

    def main(self, argv):
        # Parse args once to find version
        parser = self.get_base_parser(argv)
        (options, args) = parser.parse_known_args(argv)
        self._setup_logging(options.debug)

        # build available subcommands based on version
        api_version = options.murano_api_version
        subcommand_parser = self.get_subcommand_parser(api_version, argv)
        self.parser = subcommand_parser

        keystone_session = None
        keystone_auth = None

        # Handle top-level --help/-h before attempting to parse
        # a command off the command line.
        if (not args and options.help) or not argv:
            self.do_help(options)
            return 0

        # Parse args again and call whatever callback was selected.
        args = subcommand_parser.parse_args(argv)

        # Short-circuit and deal with help command right away.
        if args.func == self.do_help:
            self.do_help(args)
            return 0
        elif args.func == self.do_bash_completion:
            self.do_bash_completion(args)
            return 0

        if not args.os_username and not args.os_auth_token:
            raise exc.CommandError("You must provide a username via"
                                   " either --os-username or env[OS_USERNAME]"
                                   " or a token via --os-auth-token or"
                                   " env[OS_AUTH_TOKEN]")

        if args.murano_packages_service == 'glance':
            args.murano_packages_service = 'glare'

        if args.os_no_client_auth:
            if not args.murano_url:
                raise exc.CommandError(
                    "If you specify --os-no-client-auth"
                    " you must also specify a Murano API URL"
                    " via either --murano-url or env[MURANO_URL]")
            if (not args.glare_url and
                    args.murano_packages_service == 'glare'):
                raise exc.CommandError(
                    "If you specify --os-no-client-auth and"
                    " set murano-packages-service to 'glare'"
                    " you must also specify a glare API URL"
                    " via either --glare-url or env[GLARE_API]")
            if (not any([args.os_tenant_id, args.os_project_id]) and
                    args.murano_packages_service == 'glare'):
                # TODO(kzaitsev): see if we can use project's name here
                # NOTE(kzaitsev): glare v0.1 needs project_id to operate
                # correctly
                raise exc.CommandError(
                    "If you specify --os-no-client-auth and"
                    " set murano-packages-service to 'glare'"
                    " you must also specify your project's id"
                    " via either --os-project-id or env[OS_PROJECT_ID] or"
                    " --os-tenant-id or env[OS_TENANT_ID]")

        else:
            # Tenant name or ID is needed to make keystoneclient retrieve a
            # service catalog, it's not required if os_no_client_auth is
            # specified, neither is the auth URL.
            if not any([args.os_tenant_name, args.os_tenant_id,
                        args.os_project_id, args.os_project_name]):
                raise exc.CommandError("You must provide a project name or"
                                       " project id via --os-project-name,"
                                       " --os-project-id, env[OS_PROJECT_ID]"
                                       " or env[OS_PROJECT_NAME]. You may"
                                       " use os-project and os-tenant"
                                       " interchangeably.")
            if not args.os_auth_url:
                raise exc.CommandError("You must provide an auth url via"
                                       " either --os-auth-url or via"
                                       " env[OS_AUTH_URL]")

        endpoint_type = args.os_endpoint_type or 'publicURL'
        endpoint = args.murano_url
        glance_endpoint = args.glance_url

        if args.os_no_client_auth:
            # Authenticate through murano, don't use session
            kwargs = {
                'username': args.os_username,
                'password': args.os_password,
                'auth_token': args.os_auth_token,
                'auth_url': args.os_auth_url,
                'token': args.os_auth_token,
                'insecure': args.insecure,
                'timeout': args.api_timeout,
                'tenant': args.os_project_id or args.os_tenant_id,
            }
            glance_kwargs = kwargs.copy()

            if args.os_region_name:
                kwargs['region_name'] = args.os_region_name
                glance_kwargs['region_name'] = args.os_region_name
        else:
            # Create a keystone session and keystone auth
            keystone_session = ksession.Session.load_from_cli_options(args)

            args.os_project_name = args.os_project_name or args.os_tenant_name
            args.os_project_id = args.os_project_id or args.os_tenant_id

            # make args compatible with DefaultCLI/AuthCLI
            args.os_token = args.os_auth_token
            args.os_endpoint = ''
            # avoid password prompt if no password given
            args.os_password = args.os_password or '<no password>'
            (v2_auth_url, v3_auth_url) = self._discover_auth_versions(
                keystone_session, args.os_auth_url)
            if v3_auth_url:
                if (not args.os_user_domain_id and
                        not args.os_user_domain_name):
                    args.os_user_domain_name = 'default'

                if (not args.os_project_domain_id and
                        not args.os_project_domain_name):
                    args.os_project_domain_name = 'default'

            keystone_auth = AuthCLI.load_from_argparse_arguments(args)

            service_type = args.os_service_type or 'application-catalog'

            if not endpoint:
                endpoint = keystone_auth.get_endpoint(
                    keystone_session,
                    service_type=service_type,
                    interface=endpoint_type,
                    region_name=args.os_region_name)

            kwargs = {
                'session': keystone_session,
                'auth': keystone_auth,
                'service_type': service_type,
                'region_name': args.os_region_name,
            }
            glance_kwargs = kwargs.copy()

            # glance doesn't need endpoint_type
            kwargs['endpoint_type'] = endpoint_type
            kwargs['tenant'] = keystone_auth.get_project_id(keystone_session)

        if args.api_timeout:
            kwargs['timeout'] = args.api_timeout

        if not glance_endpoint:
            try:
                glance_endpoint = keystone_auth.get_endpoint(
                    keystone_session,
                    service_type='image',
                    interface=endpoint_type,
                    region_name=args.os_region_name)
            except Exception:
                pass

        glance_client = None
        if glance_endpoint:
            try:
                # TODO(starodubcevna): switch back to glance APIv2 when it will
                # be ready for use.
                glance_client = glanceclient.Client(
                    '1', glance_endpoint, **glance_kwargs)
            except Exception:
                pass
        if glance_client:
            kwargs['glance_client'] = glance_client
        else:
            logger.warning("Could not initialize glance client. "
                           "Image creation will be unavailable.")
            kwargs['glance_client'] = None

        if args.murano_packages_service == 'glare':
            glare_endpoint = args.glare_url

            if not glare_endpoint:
                # no glare_endpoint and we requested to store packages in glare
                # let's check keystone
                try:
                    glare_endpoint = keystone_auth.get_endpoint(
                        keystone_session,
                        service_type='artifact',
                        interface=endpoint_type,
                        region_name=args.os_region_name)
                except Exception:
                    raise exc.CommandError(
                        "You set murano-packages-service to {}"
                        " but there is not 'artifact' endpoint in keystone"
                        " Either register one or specify endpoint "
                        " via either --glare-url or env[GLARE_API]".format(
                            args.murano_packages_service))

            auth_token = \
                args.os_auth_token or keystone_auth.get_token(keystone_session)

            artifacts_client = art_client.Client(endpoint=glare_endpoint,
                                                 type_name='murano',
                                                 type_version=1,
                                                 token=auth_token,
                                                 insecure=args.insecure)
            kwargs['artifacts_client'] = artifacts_client

        client = murano_client.Client(api_version, endpoint, **kwargs)

        args.func(client, args)

    def do_bash_completion(self, args):
        """Prints all of the commands and options to stdout."""
        commands = set()
        options = set()
        for sc_str, sc in self.subcommands.items():
            commands.add(sc_str)
            for option in list(sc._optionals._option_string_actions):
                options.add(option)

        commands.remove('bash-completion')
        print(' '.join(commands | options))

    @utils.arg('command', metavar='<subcommand>', nargs='?',
               help='Display help for <subcommand>')
    def do_help(self, args):
        """Display help about this program or one of its subcommands."""
        if getattr(args, 'command', None):
            if args.command in self.subcommands:
                self.subcommands[args.command].print_help()
            else:
                msg = "'%s' is not a valid subcommand"
                raise exc.CommandError(msg % args.command)
        else:
            self.parser.print_help()


class HelpFormatter(argparse.HelpFormatter):
    def start_section(self, heading):
        # Title-case the headings
        heading = '%s%s' % (heading[0].upper(), heading[1:])
        super(HelpFormatter, self).start_section(heading)


def main(args=sys.argv[1:]):
    try:
        MuranoShell().main(args)

    except KeyboardInterrupt:
        print('... terminating murano client', file=sys.stderr)
        sys.exit(1)
    except Exception as e:
        if '--debug' in args or '-d' in args:
            raise
        else:
            print(encodeutils.safe_encode(str(e)), file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()
